from pathlib import Path
from typing import Union

from epl.protobuf.v1 import stac_pb2
from nsl.stac.destinations.base import BaseDestination
from nsl.stac.enum import AssetType
from nsl.stac.utils import download_asset


class MemoryDestination(BaseDestination):
    type = 'memory'
    save_directory: Path

    def __init__(self, asset_type: AssetType = AssetType.GEOTIFF, save_directory: Union[Path, str] = '/'):
        super().__init__(asset_type=asset_type)
        self.save_directory = Path(save_directory)

    def deliver(self, nsl_id: str, sub_id: str, stac_item: stac_pb2.StacItem):
        try:
            file_path = self.save_directory.joinpath(self.file_name(stac_item))
            download_asset(asset=self.asset(stac_item),
                           from_bucket=True,
                           save_filename=str(file_path))
            return None
        except BaseException as err:
            print(f'ERROR: failed downloading asset for {stac_item.id}:\n{err}')
            raise err

    def __json__(self) -> dict:
        return dict(**super().__json__(), save_directory=str(self.save_directory))



import os

from pathlib import Path
from typing import Optional, Union

import boto3

from epl.protobuf.v1 import stac_pb2
from nsl.stac.enum import AssetType
from nsl.stac.destinations.base import BaseDestination

AWS_ACCESS_KEY_ID = os.getenv('AWS_ACCESS_KEY_ID')
AWS_SECRET_ACCESS_KEY = os.getenv('AWS_SECRET_ACCESS_KEY')


class AWSDestination(BaseDestination):
    type = 'aws'

    _client = None
    save_directory: Path
    role_arn: str
    bucket: str
    region: str

    def __init__(self,
                 role_arn: str,
                 bucket: str,
                 region: str,
                 asset_type: AssetType = AssetType.GEOTIFF,
                 save_directory: Union[Path, str] = '/'):
        super().__init__(asset_type=asset_type)
        self.save_directory = Path(save_directory)
        self.role_arn = role_arn
        self.bucket = bucket
        self.region = region

    def deliver(self, nsl_id: str, sub_id: str, stac_item: stac_pb2.StacItem):
        try:
            # TODO: in the future, check a local cache for this asset to avoid multiple downloads
            self.target_obj(stac_item).upload_fileobj(self.src_blob(stac_item).open('rb'))
            return None
        except BaseException as err:
            print(f'ERROR: failed to transfer asset {stac_item.id}:\n{err}')
            raise err

    def __json__(self) -> dict:
        return dict(**super().__json__(),
                    role_arn=self.role_arn,
                    bucket=self.bucket,
                    region=self.region,
                    save_directory=str(self.save_directory))

    def target_obj(self, stac_item: stac_pb2.StacItem) -> Optional[object]:
        return self.s3.Object(self.bucket, self.blob_path(stac_item, self.save_directory))

    @property
    def s3(self):
        if self._client is None:
            client = boto3\
                .Session(aws_access_key_id=AWS_ACCESS_KEY_ID, aws_secret_access_key=AWS_SECRET_ACCESS_KEY)\
                .client('sts')
            assumed_role = client.assume_role(RoleArn=self.role_arn, RoleSessionName='nearspacelabs')
            credentials = assumed_role['Credentials']
            self._client = boto3.resource('s3',
                                          aws_access_key_id=credentials['AccessKeyId'],
                                          aws_secret_access_key=credentials['SecretAccessKey'],
                                          aws_session_token=credentials['SessionToken'])
        return self._client



# Copyright 2019-20 Near Space Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# for additional information, contact:
#   info@nearspacelabs.com
import uuid

import requests

from typing import Iterator, List, Optional, Tuple
from warnings import warn

from epl.protobuf.v1 import stac_pb2

from nsl.stac import AUTH0_TENANT, bearer_auth, stac_service as stac_singleton, utils, TimestampFilter
from nsl.stac.destinations import BaseDestination, MemoryDestination
from nsl.stac.subscription import Subscription
from nsl.stac.utils import item_region


class NSLClient:
    def __init__(self, nsl_only=True, nsl_id=None, profile_name=None):
        """
        Create a client connection to a gRPC STAC service. nsl_only limits all queries to only return data from Near
        Space Labs.
        :param nsl_only:
        """
        self._stac_service = stac_singleton
        self._nsl_only = nsl_only
        if profile_name:
            nsl_id = bearer_auth._get_auth_info(profile_name=profile_name).nsl_id
        if nsl_id:
            bearer_auth._default_nsl_id = nsl_id

    @property
    def default_nsl_id(self):
        """
        if you don't set the nsl_id for each stac request, then this nsl_id is the default choice.
        if you set this default value you must make sure that the nsl_id has already been 'set' by calling `set_credentials`
        :return:
        """
        return bearer_auth.default_nsl_id

    def set_credentials(self, nsl_id: str, nsl_secret: str):
        """
        Set nsl_id and secret for use in querying metadata and downloading imagery
        :param nsl_id:
        :param nsl_secret:
        """
        bearer_auth.set_credentials(nsl_id=nsl_id, nsl_secret=nsl_secret)

    def update_service_url(self, stac_service_url):
        """
        update the stac service address
        :param stac_service_url: localhost:8080, 34.34.34.34:9000, http://api.nearspacelabs.net:9090, etc
        :return:
        """
        self._stac_service.update_service_url(stac_service_url=stac_service_url)

    def search_one(self,
                   stac_request: stac_pb2.StacRequest,
                   timeout=15,
                   nsl_id: str = None,
                   profile_name: str = None,
                   correlation_id: str = None) -> stac_pb2.StacItem:
        """
        search for one item from the db that matches the stac request
        :param timeout: timeout for request
        :param stac_request: StacRequest of query parameters to filter by
        :param nsl_id: ADVANCED ONLY. Only necessary if more than one nsl_id and nsl_secret have been defined with
        set_credentials method.  Specify nsl_id to use. if NSL_ID and NSL_SECRET environment variables not set must use
        NSLClient object's set_credentials to set credentials
        :param profile_name: if a ~/.nsl/credentials file exists, you can override the [default] credential usage, by
        using a different profile name
        :param correlation_id: is a unique identifier that is added to the very first interaction (incoming request)
        to identify the context and is passed to all components that are involved in the transaction flow
        :return: StacItem
        """
        # limit to only search Near Space Labs SWIFT data
        if self._nsl_only:
            stac_request.mission_enum = stac_pb2.SWIFT

        metadata = self._grpc_headers(nsl_id, profile_name, correlation_id)
        return self._stac_service.stub.SearchOneItem(stac_request, timeout=timeout, metadata=metadata)

    def count(self,
              stac_request: stac_pb2.StacRequest,
              timeout=15,
              nsl_id: str = None,
              profile_name: str = None,
              correlation_id: str = None) -> int:
        """
        count all the items in the database that match the stac request
        :param timeout: timeout for request
        :param stac_request: StacRequest query parameters to apply to count method (limit ignored)
        :param nsl_id: ADVANCED ONLY. Only necessary if more than one nsl_id and nsl_secret have been defined with
        set_credentials method.  Specify nsl_id to use. if NSL_ID and NSL_SECRET environment variables not set must use
        NSLClient object's set_credentials to set credentials
        :param profile_name: if a ~/.nsl/credentials file exists, you can override the [default] credential usage, by
        using a different profile name
        :param correlation_id: is a unique identifier that is added to the very first interaction (incoming request)
        to identify the context and is passed to all components that are involved in the transaction flow
        :return: int
        """
        # limit to only search Near Space Labs SWIFT data
        if self._nsl_only:
            stac_request.mission_enum = stac_pb2.SWIFT

        metadata = self._grpc_headers(nsl_id, profile_name, correlation_id)
        db_result = self._stac_service.stub.CountItems(stac_request, timeout=timeout, metadata=metadata)
        if db_result.status:
            # print db_result
            print(db_result.status)
        return db_result.count

    def search(self,
               stac_request: stac_pb2.StacRequest,
               timeout=15,
               nsl_id: str = None,
               profile_name: str = None,
               auto_paginate: bool = False,
               only_accessible: bool = False,
               page_size: int = 50,
               correlation_id: str = None) -> Iterator[stac_pb2.StacItem]:
        """
        search for stac items by using StacRequest. return a stream of StacItems
        :param timeout: timeout for request
        :param stac_request: StacRequest of query parameters to filter by
        :param nsl_id: ADVANCED ONLY. Only necessary if more than one nsl_id and nsl_secret have been defined with
        set_credentials method.  Specify nsl_id to use. if NSL_ID and NSL_SECRET environment variables not set must use
        NSLClient object's set_credentials to set credentials
        :param profile_name: if a ~/.nsl/credentials file exists, you can override the [default] credential usage, by
        using a different profile name
        :param auto_paginate:
            - if specified, this will automatically paginate and yield all received StacItems.
            - If `stac_request.limit` is specified, only the that amount of StacItems will be yielded.
            - If `stac_request.offset` is specified, pagination will begin at that `offset`.
            - If set to `False` (the default), `stac_request.limit` and `stac_request.offset` can be used to manually
                page through StacItems.
        :param only_accessible: limits results to only StacItems downloadable by your level of sample/paid access
        :param page_size: how many results to page at a time
        :return: stream of StacItems
        """
        for item in self._search_all(stac_request,
                                     timeout,
                                     nsl_id=nsl_id,
                                     profile_name=profile_name,
                                     auto_paginate=auto_paginate,
                                     page_size=page_size,
                                     correlation_id=correlation_id):
            if not only_accessible or \
                    bearer_auth.is_valid_for(item_region(item), nsl_id=nsl_id, profile_name=profile_name):
                yield item

    def search_collections(self,
                           collection_request: stac_pb2.CollectionRequest,
                           timeout=15,
                           nsl_id: str = None,
                           profile_name: str = None,
                           correlation_id: str = None) -> Iterator[stac_pb2.Collection]:

        metadata = self._grpc_headers(nsl_id, profile_name, correlation_id)
        for item in self._stac_service.stub.SearchCollections(collection_request, timeout=timeout, metadata=metadata):
            yield item

    def subscribe(self,
                  stac_request: stac_pb2.StacRequest,
                  destination: BaseDestination,
                  nsl_id: str = None,
                  profile_name: str = None,
                  is_active=True) -> str:
        """
        Creates a subscription to a `StacRequest`, to deliver matching `StacItem`s to a `BaseDestination`.
        """
        assert stac_request.updated == TimestampFilter(), \
            "cannot subscribe to StacRequests with a set `updated` timestamp filter"
        assert not (isinstance(destination, MemoryDestination) or destination.__class__ == BaseDestination), \
            "cannot create subscriptions that deliver to `BaseDestination`s or `MemoryDestination`s"

        if self._nsl_only:
            stac_request.mission_enum = stac_pb2.SWIFT
        res = requests.post(f'{AUTH0_TENANT}/subscription',
                            headers=self._json_headers(nsl_id, profile_name),
                            json=dict(stac_request=utils.stac_request_to_b64(stac_request),
                                      destination=destination.to_json_str(),
                                      is_active=is_active))

        NSLClient._handle_json_response(res, 201)
        sub_id = res.json()['sub_id']
        print(f'created subscription with id: {sub_id}')
        return sub_id

    def resubscribe(self, sub_id: str, nsl_id: str = None, profile_name: str = None):
        """Reactivates a subscription with the given `sub_id`."""
        res = requests.put(f'{AUTH0_TENANT}/subscription/{sub_id}',
                           headers=self._json_headers(nsl_id, profile_name))

        NSLClient._handle_json_response(res, 200)
        print(f'reactivated subscription with id: {sub_id}')
        return

    def unsubscribe(self, sub_id: str, nsl_id: str = None, profile_name: str = None):
        """Deactivates a subscription with the given `sub_id`."""
        res = requests.delete(f'{AUTH0_TENANT}/subscription/{sub_id}',
                              headers=self._json_headers(nsl_id, profile_name))

        NSLClient._handle_json_response(res, 202)
        print(f'deactivated subscription with id: {sub_id}')
        return

    def subscriptions(self, nsl_id: str = None, profile_name: str = None) -> List[Subscription]:
        """Fetches all subscriptions."""
        res = requests.get(f'{AUTH0_TENANT}/subscription',
                           headers=self._json_headers(nsl_id, profile_name))

        NSLClient._handle_json_response(res, 200)
        return list(Subscription(response_dict) for response_dict in res.json()['results'])

    def _search_all(self,
                    stac_request: stac_pb2.StacRequest,
                    timeout=15,
                    nsl_id: str = None,
                    profile_name: str = None,
                    auto_paginate: bool = False,
                    page_size: int = 50,
                    correlation_id: str = None) -> Iterator[stac_pb2.StacItem]:
        # limit to only search Near Space Labs SWIFT data
        if self._nsl_only:
            stac_request.mission_enum = stac_pb2.SWIFT

        if not auto_paginate:
            metadata = self._grpc_headers(nsl_id, profile_name, correlation_id)
            for item in self._stac_service.stub.SearchItems(stac_request, timeout=timeout, metadata=metadata):
                if not item.id:
                    warn(f"STAC item missing STAC id: \n{item};\n ending search")
                    return
                else:
                    yield item
        else:
            original_limit = stac_request.limit if stac_request.limit > 0 else None
            offset = stac_request.offset
            count = 0

            stac_request.limit = page_size if original_limit is None else max(original_limit, page_size)
            items = list(self._search_all(stac_request, timeout=timeout,
                                          nsl_id=nsl_id, profile_name=profile_name,
                                          page_size=page_size, correlation_id=correlation_id))
            while len(items) > 0:
                for item in items:
                    if original_limit is None or (original_limit is not None and count < original_limit):
                        yield item
                        count += 1
                    if original_limit is not None and count >= original_limit:
                        break

                if original_limit is not None and count >= original_limit:
                    break

                stac_request.offset += len(items)
                items = list(self._search_all(stac_request, timeout=timeout,
                                              nsl_id=nsl_id, profile_name=profile_name,
                                              page_size=page_size, correlation_id=correlation_id))

            stac_request.offset = offset
            stac_request.limit = original_limit if original_limit is not None else 0

    def _json_headers(self,
                      nsl_id: str = None,
                      profile_name: str = None,
                      correlation_id: str = None) -> dict:
        headers = {k: v for (k, v) in self._grpc_headers(nsl_id, profile_name, correlation_id)}
        return {'content-type': 'application/json', **headers}

    def _grpc_headers(self,
                      nsl_id: str = None,
                      profile_name: str = None,
                      correlation_id: str = None) -> Tuple[Tuple[str, str], ...]:
        correlation_id = str(uuid.uuid4()) if correlation_id is None else correlation_id
        return (('x-correlation-id', correlation_id),
                ('authorization', bearer_auth.auth_header(nsl_id=nsl_id, profile_name=profile_name)))

    @staticmethod
    def _handle_json_response(res, status_code: int):
        if res.status_code != status_code:
            raise requests.exceptions.RequestException(f'non-nominal status code: {res.status_code}')
        elif len(res.content) == 0:
            # then if response is empty, HTTPResponse method for read returns b"" which will be zero in length
            raise requests.exceptions.RequestException("empty authentication return. notify nsl of error")



from .base import BaseDestination, DestinationDecoder
from .aws import AWSDestination
from .gcp import GCPDestination
from .memory import MemoryDestination

__all__ = ['BaseDestination', 'DestinationDecoder',
           'AWSDestination', 'GCPDestination', 'MemoryDestination']



# Copyright 2019-20 Near Space Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# for additional information, contact:
#   info@nearspacelabs.com

import abc
import base64
import os
import re
import json
import math
import time
import warnings
import logging

import grpc
import requests

from dataclasses import dataclass
from pathlib import Path
from random import randint
from typing import Dict, Optional, Set, Tuple

from google.auth.exceptions import DefaultCredentialsError
from google.cloud import storage as gcp_storage
from google.oauth2 import service_account
from tenacity import retry, stop_after_delay, wait_fixed

from epl.protobuf.v1 import stac_service_pb2_grpc
from epl.protobuf.v1.geometry_pb2 import GeometryData, ProjectionData, EnvelopeData
from epl.protobuf.v1.query_pb2 import TimestampFilter, FloatFilter, StringFilter, UInt32Filter
from epl.protobuf.v1.stac_pb2 import StacRequest, StacItem, Asset, Collection, CollectionRequest, Eo, EoRequest, \
    LandsatRequest, Mosaic, MosaicRequest, DatetimeRange, View, ViewRequest, Extent, Interval, Provider

__all__ = [
    'bearer_auth', 'gcs_storage_client', 'stac_service', 'url_to_channel',
    'CollectionRequest', 'EoRequest', 'StacRequest', 'LandsatRequest', 'MosaicRequest', 'ViewRequest',
    'Collection', 'Eo', 'StacItem', 'Mosaic', 'View', 'Asset',
    'GeometryData', 'ProjectionData', 'EnvelopeData', 'FloatFilter', 'TimestampFilter', 'StringFilter', 'UInt32Filter',
    'DatetimeRange', 'Extent', 'Interval', 'Provider',
    'AUTH0_TENANT', 'API_AUDIENCE', 'ISSUER', 'STAC_SERVICE', 'AuthInfo',
]

CLOUD_PROJECT = os.getenv("CLOUD_PROJECT")
GOOGLE_APPLICATION_CREDENTIALS = os.getenv("GOOGLE_APPLICATION_CREDENTIALS")
SERVICE_ACCOUNT_DETAILS = os.getenv("SERVICE_ACCOUNT_DETAILS")

NSL_ID = os.getenv("NSL_ID")
NSL_SECRET = os.getenv("NSL_SECRET")
# if an application uses this package, this singleton pattern of using an __init__.py file means that the package
# and all the network calls will be made immediately upon application start. In some cases a virtual machine might
# be able to start an application before it has network access.
NSL_NETWORK_DELAY = int(os.getenv("NSL_NETWORK_DELAY", 0))

# URL of the OAuth service
AUTH0_TENANT = os.getenv('AUTH0_TENANT', 'https://api.nearspacelabs.net')
# Name of the API for which we issue tokens
API_AUDIENCE = os.getenv('API_AUDIENCE', 'https://api.nearspacelabs.com')
# Name of the service responsible for issuing NSL tokens
ISSUER = 'https://api.nearspacelabs.net/'

TOKEN_REFRESH_THRESHOLD = 60  # seconds

MAX_GRPC_ATTEMPTS = int(os.getenv('MAX_ATTEMPTS', 4))
INIT_BACKOFF_MS = int(os.getenv('INIT_BACKOFF_MS', 4))
MAX_BACKOFF_MS = int(os.getenv('MAX_BACKOFF_MS', 4))
MULTIPLIER = int(os.getenv('MULTIPLIER', 4))

STAC_SERVICE_HOST = os.getenv('STAC_SERVICE_HOST', 'api.nearspacelabs.net')
STAC_SERVICE = os.getenv('STAC_SERVICE', f'{STAC_SERVICE_HOST}:9090')
BYTES_IN_MB = 1024 * 1024
# at this point only allowing 10 MB or smaller messages
MESSAGE_SIZE_MB = int(os.getenv('MESSAGE_SIZE_MB', 10))
GRPC_CHANNEL_OPTIONS = [('grpc.max_message_length', MESSAGE_SIZE_MB * BYTES_IN_MB),
                        ('grpc.max_receive_message_length', MESSAGE_SIZE_MB * BYTES_IN_MB)]

# TODO prep for ip v6
IP_REGEX = re.compile(r"[\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}")
# DEFAULT Insecure until we have a https service
INSECURE = True
NSL_CREDENTIALS = Path(Path.home(), '.nsl', 'credentials')

logger = logging.getLogger()


class SleepingPolicy(abc.ABC):
    @abc.abstractmethod
    def sleep(self, try_i: int):
        """
        How long to sleep in milliseconds.
        :param try_i: the number of retry (starting from zero)
        """
        assert try_i >= 0


# Retry mechanism
# https://github.com/grpc/grpc/issues/19514#issuecomment-531700657
class ExponentialBackoff(SleepingPolicy):
    def __init__(self, *, init_backoff_ms: int, max_backoff_ms: int, multiplier: int):
        self.init_backoff = randint(0, init_backoff_ms)
        self.max_backoff = max_backoff_ms
        self.multiplier = multiplier

    def sleep(self, try_i: int):
        sleep_range = min(self.init_backoff * self.multiplier ** try_i, self.max_backoff)
        sleep_ms = randint(0, sleep_range)
        logger.debug(f"Sleeping for {sleep_ms}")
        time.sleep(sleep_ms / 1000)


class RetryOnRpcErrorClientInterceptor(grpc.UnaryUnaryClientInterceptor, grpc.StreamUnaryClientInterceptor):
    def __init__(self,
                 *,
                 max_attempts: int,
                 sleeping_policy: SleepingPolicy,
                 status_for_retry: Optional[Tuple[grpc.StatusCode]] = None):
        self.max_attempts = max_attempts
        self.sleeping_policy = sleeping_policy
        self.status_for_retry = status_for_retry

    def _intercept_call(self, continuation, client_call_details, request_or_iterator):
        for try_i in range(self.max_attempts):
            response = continuation(client_call_details, request_or_iterator)

            if isinstance(response, grpc.RpcError):

                # Return if it was last attempt
                if try_i == (self.max_attempts - 1):
                    return response

                # If status code is not in retryable status codes
                if self.status_for_retry and response.code() not in self.status_for_retry:
                    return response

                self.sleeping_policy.sleep(try_i)
            else:
                return response

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        return self._intercept_call(continuation, client_call_details, request_iterator)


interceptors = (
    RetryOnRpcErrorClientInterceptor(
        max_attempts=MAX_GRPC_ATTEMPTS,
        sleeping_policy=ExponentialBackoff(init_backoff_ms=INIT_BACKOFF_MS,
                                           max_backoff_ms=MAX_BACKOFF_MS,
                                           multiplier=MULTIPLIER),
        status_for_retry=(grpc.StatusCode.UNAVAILABLE,),
    ),
)


def url_to_channel(stac_service_url):
    if stac_service_url.startswith("localhost") or IP_REGEX.match(stac_service_url) or \
            "." not in stac_service_url or stac_service_url.startswith("http://") or INSECURE:
        stac_service_url = stac_service_url.strip("http://")
        channel = grpc.insecure_channel(stac_service_url, options=GRPC_CHANNEL_OPTIONS)
    else:
        stac_service_url = stac_service_url.strip("https://")
        channel_credentials = grpc.ssl_channel_credentials()
        channel = grpc.secure_channel(stac_service_url,
                                      credentials=channel_credentials,
                                      options=GRPC_CHANNEL_OPTIONS)

    return grpc.intercept_channel(channel, *interceptors)


class __GCSStorageClient:
    _client = None

    @property
    def client(self):
        if self._client is not None:
            return self._client

        if SERVICE_ACCOUNT_DETAILS:
            details = json.loads(SERVICE_ACCOUNT_DETAILS)
            creds = service_account.Credentials.from_service_account_info(details)
            client = gcp_storage.Client(project=CLOUD_PROJECT, credentials=creds)
        elif GOOGLE_APPLICATION_CREDENTIALS:
            creds = service_account.Credentials.from_service_account_file(GOOGLE_APPLICATION_CREDENTIALS)
            client = gcp_storage.Client(project=CLOUD_PROJECT, credentials=creds)
        else:
            try:
                # https://github.com/googleapis/google-auth-library-python/issues/271#issuecomment-400186626
                warnings.filterwarnings("ignore", "Your application has authenticated using end user credentials")
                client = gcp_storage.Client(project="")
            except DefaultCredentialsError:
                client = None
        self._client = client
        return self._client


def _generate_grpc_channel(stac_service_url=None):
    # TODO host env should include http:// so we can just see if it's https or http

    if stac_service_url is None:
        stac_service_url = STAC_SERVICE

    channel = url_to_channel(stac_service_url)

    print("nsl client connecting to stac service at: {}\n".format(stac_service_url))

    return channel, stac_service_pb2_grpc.StacServiceStub(channel)


class __StacServiceStub(object):
    def __init__(self):
        channel, stub = _generate_grpc_channel()
        self._channel = channel
        self._stub = stub

    @property
    def channel(self):
        return self._channel

    @property
    def stub(self):
        return self._stub

    def set_channel(self, channel):
        """
        This allows you to override the channel created on init, with another channel. This might be needed if multiple
        libraries are using the same channel, or if multi-threading.
        :param channel:
        :return:
        """
        self._stub = stac_service_pb2_grpc.StacServiceStub(channel)
        self._channel = channel

    def update_service_url(self, stac_service_url):
        """allows you to update your stac service address"""
        self._channel, self._stub = _generate_grpc_channel(stac_service_url)


@dataclass
class Contract:
    balance: int
    region: int
    type: str

    @staticmethod
    def from_jwt(token: str):
        payload = json.loads(base64.b64decode(token.split('.')[1] + '=='))
        contract = payload[f'{API_AUDIENCE}/contract']
        return Contract(balance=contract['balance'], region=contract['region'], type=contract['type'])

    def is_valid_for(self, region: str) -> bool:
        if self.region == 1 or region == 'REGION_0' or region == 'SAMPLES':
            return True

        if region == "REGION_1":
            masked = self.region & 2
        elif region == "REGION_2":
            masked = self.region & 4
        elif region == "REGION_3":
            masked = self.region & 8
        elif region == "REGION_4":
            masked = self.region & 16
        elif region == "REGION_5":
            masked = self.region & 32
        elif region == "REGION_6":
            masked = self.region & 64
        elif region == "REGION_7":
            masked = self.region & 128
        else:
            return False
        return masked != 0 and masked <= self.region


class AuthInfo:
    nsl_id: str = None
    nsl_secret: str = None
    token: str = None
    expiry: float = 0
    skip_authorization: bool = False
    contract: Contract

    def __init__(self, nsl_id: str, nsl_secret: str):
        if not nsl_id or not nsl_secret:
            raise ValueError("nsl_id and nsl_secret must be non-zero length strings")
        self.nsl_id = nsl_id
        self.nsl_secret = nsl_secret

    # this only retries if there's a timeout error
    @retry(reraise=True, stop=stop_after_delay(3), wait=wait_fixed(0.5))
    def authorize(self):
        if self.skip_authorization:
            return

        expiry, token = AuthInfo.get_token_client_credentials(self.nsl_id, self.nsl_secret)
        self.expiry = expiry
        self.token = token
        self.contract = Contract.from_jwt(token)

    @property
    def permissions(self) -> Set[str]:
        _, payload, _ = self.token.split('.')
        payload = json.loads(base64.b64decode(payload + '==='))
        return {permission for permission in payload.get('permissions', set())}

    @staticmethod
    def get_token_client_credentials(nsl_id: str, nsl_secret: str,
                                     auth_url=f"{AUTH0_TENANT}/oauth/token",
                                     grant_type: str = 'client_credentials',
                                     additional_headers: dict = None):
        print(f"attempting NSL authentication against {auth_url}...")
        now = time.time()

        if additional_headers is None:
            additional_headers = dict()

        headers = {'content-type': 'application/json', **additional_headers}
        post_body = {
            'client_id': nsl_id,
            'client_secret': nsl_secret,
            'audience': API_AUDIENCE,
            'grant_type': grant_type,
        }

        res = requests.post(auth_url, json=post_body, headers=headers)

        if res.status_code != 200 and res.status_code != 201:
            # evaluate codes first.
            message = f"authentication failed with code '{res.status_code}' and reason '{res.reason}'"
            raise requests.exceptions.RequestException(message)
        elif len(res.content) == 0:
            # then if response is empty, HTTPResponse method for read returns b"" which will be zero in length
            raise requests.exceptions.RequestException("empty authentication return. notify nsl of error")

        print(f"successfully authenticated with NSL_ID: `{nsl_id}`")
        res_json = res.json()
        expiry = now + int(res_json['expires_in'])
        token = res_json['access_token']
        return expiry, token


class __BearerAuth:
    _auth_info_map: Dict[str, AuthInfo] = {}
    _profile_map: Dict[str, str] = {}
    _default_nsl_id = None

    def __init__(self, init=False):
        if (not NSL_ID or not NSL_SECRET) and not NSL_CREDENTIALS.exists():
            warnings.warn(f"NSL_ID and NSL_SECRET environment variables not set, and {NSL_CREDENTIALS} does not exist")
            return

        # if credentials exist, add them to our auth store
        if NSL_CREDENTIALS and NSL_CREDENTIALS.exists():
            for profile_name, auth_info in self.loads().items():
                self._auth_info_map[auth_info.nsl_id] = auth_info
                if profile_name == 'default':
                    self._default_nsl_id = auth_info.nsl_id
                self._profile_map[profile_name] = auth_info.nsl_id
                print(f"found NSL_ID {auth_info.nsl_id} under profile name `{profile_name}`")

        # if env vars were specified, add them as well and set them to the default
        if NSL_ID and NSL_SECRET:
            print(f"using NSL_ID {NSL_ID} specified in env var")
            self._auth_info_map[NSL_ID] = AuthInfo(nsl_id=NSL_ID, nsl_secret=NSL_SECRET)
            self._default_nsl_id = NSL_ID

        # if env vars are unset and no NSL_ID was tagged as default, use the first one available
        if self.default_nsl_id is None:
            self._default_nsl_id = list(key for key in self._auth_info_map.keys())[0]
            print(f"using NSL_ID {self.default_nsl_id}")

        if init:
            self._auth_info_map[self._default_nsl_id].authorize()

    @property
    def default_nsl_id(self):
        return self._default_nsl_id

    def auth_header(self, nsl_id: str = None, profile_name: str = None) -> str:
        auth_info = self._get_auth_info(nsl_id, profile_name)
        if not auth_info.skip_authorization and (auth_info.expiry - time.time()) < TOKEN_REFRESH_THRESHOLD:
            print(f'authorizing NSL_ID: `{auth_info.nsl_id}`')
            auth_info.authorize()
            diff_seconds = auth_info.expiry - time.time()
            ttl = round(int(math.ceil(float(diff_seconds / 60) / 10) * 10))
            print(f"will attempt re-authorization in {ttl} minutes")
        return f"Bearer {auth_info.token}"

    def get_credentials(self, nsl_id: str = None, profile_name: str = None) -> Optional[AuthInfo]:
        if profile_name is not None:
            nsl_id = self._profile_map.get(profile_name, None)
        return self._auth_info_map.get(nsl_id if nsl_id is not None else self.default_nsl_id, None)

    def set_credentials(self, nsl_id: str, nsl_secret: str, profile_name: str = None):
        if len(self._auth_info_map) == 0:
            self._default_nsl_id = nsl_id

        self._auth_info_map[nsl_id] = AuthInfo(nsl_id=nsl_id, nsl_secret=nsl_secret)
        self._auth_info_map[nsl_id].authorize()
        if profile_name is not None:
            self._profile_map[profile_name] = nsl_id

    def unset_credentials(self, profile_name: str):
        nsl_id = self._profile_map.pop(profile_name)
        delattr(self._auth_info_map, nsl_id)
        if self._default_nsl_id == nsl_id:
            if len(self._auth_info_map) == 0:
                self._default_nsl_id = None
            else:
                self._default_nsl_id = list(key for key in self._auth_info_map.keys())[0]
                print(f"using NSL_ID {self.default_nsl_id}")

    def is_valid_for(self, region: str, nsl_id: str = None, profile_name: str = None) -> bool:
        auth_info = self._get_auth_info(nsl_id=nsl_id, profile_name=profile_name)
        return auth_info.contract.is_valid_for(region)

    def loads(self) -> Dict[str, AuthInfo]:
        output = dict()
        with NSL_CREDENTIALS.open('r') as file_obj:
            lines = file_obj.readlines()
            for i, line in enumerate(lines):
                if line.startswith('['):
                    if not lines[i + 1].startswith('NSL_ID') or not lines[i + 2].startswith('NSL_SECRET'):
                        raise ValueError("credentials should be of the format:\n[named profile]\nNSL_ID={your "
                                         "nsl id}\nNSL_SECRET={your nsl secret}")
                    # for id like 'NSL_ID = all_the_id_text\n', first strip remove front whitespace and newline, and optionally the leading quote
                    # .strip(), now we now [6:] starts after 'NSL_ID' .strip()[6:], strip potential whitespace
                    # between NSL_ID and '=' with .strip()[6:].strip(), start one after equal
                    # .strip()[6:].strip()[1:], strip potential whitespace
                    # after equal .strip()[6:].strip()[1:].strip()
                    profile_name = line.strip().lstrip('[').rstrip(']')
                    nsl_id = lines[i + 1].strip()[6:].strip().strip('"')[1:].strip().strip('"')
                    nsl_secret = lines[i + 2].strip()[10:].strip().strip('"')[1:].strip().strip('"')

                    output[profile_name] = AuthInfo(nsl_id=nsl_id, nsl_secret=nsl_secret)
        return output

    def dumps(self):
        with NSL_CREDENTIALS.open('w') as file_obj:
            for profile_name, nsl_id in self._profile_map.items():
                creds = self.get_credentials(nsl_id)
                file_obj.write(f'[{profile_name}]\n')
                file_obj.write(f'NSL_ID="{creds.nsl_id}"\n')
                file_obj.write(f'NSL_SECRET="{creds.nsl_secret}"\n')
                file_obj.write('\n')
            file_obj.close()

    def _get_auth_info(self, nsl_id: str = None, profile_name: str = None) -> AuthInfo:
        if nsl_id is None and profile_name is None:
            nsl_id = self._default_nsl_id

        if nsl_id not in self._auth_info_map and profile_name not in self._profile_map:
            raise ValueError("credentials must be set by environment variables NSL_ID & NSL_SECRET, by setting a "
                             "credentials file at ~/.nsl/credentials, or by using the set_credentials method")

        if profile_name is not None:
            nsl_id = self._profile_map[profile_name]
        return self._auth_info_map[nsl_id]


time.sleep(NSL_NETWORK_DELAY)
bearer_auth = __BearerAuth()
stac_service = __StacServiceStub()
gcs_storage_client = __GCSStorageClient()



# Copyright 2019-20 Near Space Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# for additional information, contact:
#   info@nearspacelabs.com

import base64
import os
import datetime
import http.client
import re
from urllib.parse import urlparse
from typing import List, IO, Union, Dict, Any, Optional
from warnings import warn

import boto3
import botocore
import botocore.exceptions
import botocore.client
from google.cloud import storage
from google.protobuf import timestamp_pb2, duration_pb2
from tenacity import retry, stop_after_delay, wait_fixed

from epl.protobuf.v1.stac_pb2 import epl_dot_protobuf_dot_v1_dot_query__pb2 as query
from nsl.stac import gcs_storage_client, bearer_auth, \
    StacItem, StacRequest, Asset, TimestampFilter, DatetimeRange, Eo, FloatFilter, enum
from nsl.stac.enum import Band, CloudPlatform, FilterRelationship, SortDirection, AssetType

DEFAULT_RGB = [Band.RED, Band.GREEN, Band.BLUE, Band.NIR]
RASTER_TYPES = [AssetType.CO_GEOTIFF, AssetType.GEOTIFF, AssetType.MRF]
UNSUPPORTED_TIME_FILTERS = [FilterRelationship.IN,
                            FilterRelationship.NOT_IN,
                            FilterRelationship.LIKE,
                            FilterRelationship.NOT_LIKE]


def get_blob_metadata(bucket: str, blob_name: str) -> storage.Blob:
    """
    get metadata/interface for one asset in google cloud storage
    :param bucket: bucket name
    :param blob_name: complete blob name of item (doesn't include bucket name)
    :return: Blob interface item
    """
    if gcs_storage_client.client is None:
        raise ValueError("GOOGLE_APPLICATION_CREDENTIALS environment variable not set")

    bucket = gcs_storage_client.client.get_bucket(bucket)
    return bucket.get_blob(blob_name=blob_name.strip('/'))


@retry(reraise=True, stop=stop_after_delay(3), wait=wait_fixed(0.5))
def download_gcs_object(bucket: str,
                        blob_name: str,
                        file_obj: IO[bytes] = None,
                        save_filename: str = "",
                        make_dir=True) -> str:
    """
    download a specific blob from Google Cloud Storage (GCS) to a file object handle
    :param make_dir: if directory doesn't exist create
    :param bucket: bucket name
    :param blob_name: the full prefix to a specific asset in GCS. Does not include bucket name
    :param file_obj: file object (or BytesIO string_buffer) where data should be written
    :param save_filename: the filename to save the file to
    :return: returns path to downloaded file if applicable
    """
    if make_dir and save_filename != "":
        path_to_create = os.path.split(save_filename)[0]
        if not os.path.exists(path_to_create):
            os.makedirs(path_to_create, exist_ok=True)

    blob = get_blob_metadata(bucket=bucket, blob_name=blob_name)

    if file_obj is not None:
        blob.download_to_file(file_obj=file_obj, client=gcs_storage_client.client)
        if "name" in file_obj.__dict__:
            save_filename = file_obj.name
        else:
            save_filename = ""
        try:
            file_obj.seek(0)
        except:
            pass

        return save_filename
    elif len(save_filename) > 0:
        with open(save_filename, "w+b") as file_obj:
            download_gcs_object(bucket, blob_name, file_obj=file_obj)
        return save_filename
    else:
        raise ValueError("must provide filename or file_obj")


def download_s3_object(bucket: str,
                       blob_name: str,
                       file_obj: IO = None,
                       save_filename: str = "",
                       requester_pays: bool = False) -> str:
    extra_args = None
    if requester_pays:
        extra_args = {'RequestPayer': 'requester'}

    s3 = boto3.client('s3')
    try:
        if file_obj is not None:
            s3.download_fileobj(Bucket=bucket, Key=blob_name, Fileobj=file_obj, ExtraArgs=extra_args)
            if "name" in file_obj.__dict__:
                save_filename = file_obj.name
            else:
                save_filename = ""
            file_obj.seek(0)

            return save_filename
        elif len(save_filename) > 0:
            s3.download_file(Bucket=bucket, Key=blob_name, Filename=save_filename, ExtraArgs=extra_args)
            return save_filename
        else:
            raise ValueError("must provide filename or file_obj")
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == "404":
            print("The object does not exist.")
        else:
            raise


def download_href_object(asset: Asset,
                         file_obj: IO = None,
                         save_filename: str = "",
                         nsl_id: str = None,
                         profile_name: str = None) -> str:
    """
    download the href of an asset
    :param asset: The asset to download
    :param file_obj: BinaryIO file object to download data into. If file_obj and save_filename and/or save_directory
    are set, then only file_obj is used
    :param save_filename: absolute or relative path filename to save asset to (must have write permissions)
    :param nsl_id: ADVANCED ONLY. Only necessary if more than one NSL_ID and NSL_SECRET have been defined with
        set_credentials method. Specify NSL_ID to use for downloading. If NSL_ID and NSL_SECRET environment variables
        are not set, you must use `NSLClient.set_credentials` to add at least one set of credentials.
    :param profile_name: ADVANCED ONLY. Only necessary if more than one NSL profile has been defined with the
        `set_credentials` method. Specifies which NSL profile to use for downloading.
    :return: returns the save_filename. if BinaryIO is not a FileIO object type, save_filename returned is an
    empty string
    """
    if not asset.href:
        raise ValueError("no href on asset")

    host = urlparse(asset.href)
    conn = http.client.HTTPConnection(host.netloc)

    headers = {}
    asset_url = host.path
    if asset.bucket_manager == "Near Space Labs":
        headers = {"authorization": bearer_auth.auth_header(nsl_id=nsl_id, profile_name=profile_name)}
        asset_url = "/download/{object}".format(object=asset.object_path)

    if len(asset.type) > 0:
        headers["content-type"] = asset.type
    conn.request(method="GET", url=asset_url, headers=headers)

    res = conn.getresponse()
    if res.status == 404:
        raise ValueError("not found error for {path}".format(path=asset.href))
    elif res.status == 403:
        raise ValueError("auth error for asset {asset}".format(asset=asset.href))
    elif res.status == 402:
        raise ValueError("not enough credits for downloading asset {asset}".format(asset=asset.href))
    elif res.status != 200:
        raise ValueError("error code {code} for asset: {asset}".format(code=res.status, asset=asset.href))

    if len(save_filename) > 0:
        with open(save_filename, mode='wb') as f:
            f.write(res.read())
    elif file_obj is not None:
        file_obj.write(res.read())
        if "name" in file_obj.__dict__:
            save_filename = file_obj.name
        else:
            save_filename = ""
        file_obj.seek(0)
    else:
        raise ValueError("must provide filename or file_obj")

    return save_filename


def download_asset(asset: Asset,
                   from_bucket: bool = False,
                   file_obj: IO[Union[Union[str, bytes], Any]] = None,
                   save_filename: str = "",
                   save_directory: str = "",
                   requester_pays: bool = False,
                   nsl_id: str = None,
                   profile_name: str = None) -> str:
    """
    download an asset. Defaults to downloading from cloud storage. save the data to a BinaryIO file object, a filename
    on your filesystem, or to a directory on your filesystem (the filename will be chosen from the basename of the
    object).
    :param requester_pays: authorize a requester pays download. this can be costly,
    so only enable it if you understand the implications.
    :param asset: The asset to download
    :param from_bucket: force the download to occur from cloud storage instead of href endpoint
    :param file_obj: BinaryIO file object to download data into. If file_obj and save_filename and/or save_directory are
     set, then only file_obj is used
    :param save_filename: absolute or relative path filename to save asset to (must have write permissions)
    :param save_directory: absolute or relative directory path to save asset in (must have write permissions). Filename
    is derived from the basename of the object_path or the href
    :param nsl_id: ADVANCED ONLY. Only necessary if more than one NSL_ID and NSL_SECRET have been defined with
        set_credentials method. Specify NSL_ID to use for downloading. If NSL_ID and NSL_SECRET environment variables
        are not set, you must use `NSLClient.set_credentials` to add at least one set of credentials.
    :param profile_name: ADVANCED ONLY. Only necessary if more than one NSL profile has been defined with the
        `set_credentials` method. Specifies which NSL profile to use for downloading.
    :return:
    """
    if len(save_directory) > 0 and file_obj is None and len(save_filename) == 0:
        if os.path.exists(save_directory):
            save_filename = os.path.join(save_directory, os.path.basename(asset.object_path))
        else:
            raise ValueError("directory 'save_directory' doesn't exist")

    if from_bucket and asset.cloud_platform == CloudPlatform.GCP:
        return download_gcs_object(bucket=asset.bucket,
                                   blob_name=asset.object_path,
                                   file_obj=file_obj,
                                   save_filename=save_filename)
    elif from_bucket and asset.cloud_platform == CloudPlatform.AWS:
        return download_s3_object(bucket=asset.bucket,
                                  blob_name=asset.object_path,
                                  file_obj=file_obj,
                                  save_filename=save_filename,
                                  requester_pays=requester_pays)
    else:
        return download_href_object(asset=asset,
                                    file_obj=file_obj,
                                    save_filename=save_filename,
                                    nsl_id=nsl_id,
                                    profile_name=profile_name)


def download_assets(stac_item: StacItem,
                    save_directory: str,
                    from_bucket: bool = False,
                    nsl_id: str = None) -> List[str]:
    """
    Download all the assets for a StacItem into a directory
    :param nsl_id: ADVANCED ONLY. Only necessary if more than one nsl_id and nsl_secret have been defined with
    set_credentials method.  Specify nsl_id to use. if NSL_ID and NSL_SECRET environment variables not set must use
        NSLClient object's set_credentials to set credentials
    :param stac_item: StacItem containing assets to download
    :param save_directory: the directory where the files should be downloaded
    :param from_bucket: force download from bucket. if set to false downloads happen from href. defaults to False
    :return:
    """
    filenames = []
    for asset_key in stac_item.assets:
        asset = stac_item.assets[asset_key]
        filenames.append(download_asset(asset=asset,
                                        from_bucket=from_bucket,
                                        save_directory=save_directory,
                                        nsl_id=nsl_id))
    return filenames


# TODO https://pypi.org/project/Deprecated/
def get_asset(stac_item: StacItem,
              asset_type: AssetType = None,
              cloud_platform: CloudPlatform = CloudPlatform.UNKNOWN_CLOUD_PLATFORM,
              eo_bands: Eo.Band = Eo.UNKNOWN_BAND,
              asset_regex: Dict = None,
              asset_key: str = None,
              b_relaxed_types: bool = False) -> Optional[Asset]:
    """
    get a protobuf object(pb) asset from a stac item pb. If your parameters are broad (say, if you used all defaults)
    this function would only return you the first asset that matches the parameters. use
    :func:`get_assets <st.stac.utils.get_assets>` to return more than one asset from a request.
    :param stac_item: stac item whose assets we want to search by parameters
    :param asset_type: an asset_type enum to return. if not defined then it is assumed to search all asset types
    :param cloud_platform: only return assets that are hosted on the cloud platform described in the cloud_platform
    field of the item. default grabs the first asset that meets all the other parameters.
    :param band: if the data has electro-optical spectrum data, define the band you want to retrieve. if the data is
    not electro-optical then don't define this parameter (defaults to UNKNOWN_BAND)
    :param asset_basename: only return asset if the basename of the object path matches this value
    :return: asset pb object
    """
    results = get_assets(stac_item, asset_type, cloud_platform, eo_bands, asset_regex, asset_key, b_relaxed_types)
    if len(results) > 1:
        raise ValueError("must be more specific in selecting your asset. if all enums are used, try using "
                         "asset_key_regex")
    elif len(results) == 1:
        return results[0]
    return None


def _asset_types_match(desired_type: enum.AssetType,
                       asset_type: enum.AssetType,
                       b_relaxed_types: bool = False) -> bool:
    if not b_relaxed_types:
        return desired_type == asset_type
    elif desired_type == enum.AssetType.TIFF:
        return asset_type == desired_type or \
               asset_type == enum.AssetType.GEOTIFF or \
               asset_type == enum.AssetType.CO_GEOTIFF
    elif desired_type == enum.AssetType.GEOTIFF:
        return asset_type == desired_type or asset_type == enum.AssetType.CO_GEOTIFF
    return asset_type == desired_type


def equals_pb(left: Asset, right: Asset):
    """
does the AssetWrap equal a protobuf Asset
    :param other:
    :return:
    """
    return left.SerializeToString() == right.SerializeToString()


# TODO https://pypi.org/project/Deprecated/
def get_assets(stac_item: StacItem,
               asset_type: enum.AssetType = None,
               cloud_platform: CloudPlatform = CloudPlatform.UNKNOWN_CLOUD_PLATFORM,
               eo_bands: Eo.Band = Eo.UNKNOWN_BAND,
               asset_regex: Dict = None,
               asset_key: str = None,
               b_relaxed_types: bool = False) -> List[Asset]:
    """
    get a generator of assets from a stac item, filtered by the parameters.
    :param stac_item: stac item whose assets we want to search by parameters
    :param band: if the data has electro optical spectrum data, define the band you want to retrieve. if the data is not
     electro optical then don't define this parameter (defaults to UNKNOWN_BAND)
    :param asset_types: a list of asset_types to seach. if not defined then it is assumed to search all asset types
    :param cloud_platform: only return assets that are hosted on the cloud platform described in the cloud_platform
    field of the item. default grabs the first asset that meets all the other parameters.
    :param asset_basename: only return asset if the basename of the object path matches this value
    :return: asset pb object
    """
    if asset_key is not None and asset_key in stac_item.assets:
        return [stac_item.assets[asset_key]]
    elif asset_key is not None and asset_key and asset_key not in stac_item.assets:
        raise ValueError("asset_key {} not found".format(asset_key))

    results = []
    for asset_key in stac_item.assets:
        current = stac_item.assets[asset_key]
        b_asset_type_match = _asset_types_match(desired_type=asset_type,
                                                asset_type=current.asset_type,
                                                b_relaxed_types=b_relaxed_types)
        if (eo_bands is not None and eo_bands != enum.Band.UNKNOWN_BAND) and \
                current.eo_bands != eo_bands:
            continue
        if (cloud_platform is not None and cloud_platform != enum.CloudPlatform.UNKNOWN_CLOUD_PLATFORM) and \
                current.cloud_platform != cloud_platform:
            continue
        if (asset_type is not None and asset_type != enum.AssetType.UNKNOWN_ASSET) and \
                not b_asset_type_match:
            continue
        if asset_regex is not None and len(asset_regex) > 0:
            b_continue = False
            for key, regex_value in asset_regex.items():
                if key == 'asset_key':
                    if not re.match(regex_value, asset_key):
                        b_continue = True
                        break
                else:
                    if not hasattr(current, key):
                        raise AttributeError("no key {0} in asset {1}".format(key, current))
                    elif not re.match(regex_value, getattr(current, key)):
                        b_continue = True
                        break

            if b_continue:
                continue

        # check that asset hasn't changed between protobuf and asset_map
        pb_asset = stac_item.assets[asset_key]
        if not equals_pb(current, pb_asset):
            raise ValueError("corrupted protobuf. Asset and AssetWrap have differing underlying protobuf")

        results.append(current)
    return results


def _asset_has_filename(asset: Asset, asset_basename):
    if os.path.basename(asset.object_path).lower() == os.path.basename(asset_basename).lower():
        return True
    return False


# TODO https://pypi.org/project/Deprecated/
def has_asset_type(stac_item: StacItem, asset_type: AssetType):
    """
    does the stac item contain the asset
    :param stac_item:
    :param asset_type:
    :return:
    """
    for asset in stac_item.assets.values():
        if asset.asset_type == asset_type:
            return True
    return False


# TODO https://pypi.org/project/Deprecated/
def has_asset(stac_item: StacItem, asset: Asset):
    """
    check whether a stac_item has a perfect match to the provided asset
    :param stac_item: stac item whose assets we're checking against asset
    :param asset: asset we're looking for in stac_item
    :return:
    """
    for test_asset in stac_item.assets.values():
        b_matches = True
        for field in test_asset.DESCRIPTOR.fields:
            if getattr(test_asset, field.name) != getattr(asset, field.name):
                b_matches = False
                break
        if b_matches:
            return b_matches
    return False


def item_region(stac_item: StacItem) -> str:
    for asset_key in stac_item.assets:
        return stac_item.assets[asset_key].object_path.split('/')[2]
    warn(f"failed to find STAC item's region: {stac_item.id}")
    return ""


def get_uri(asset: Asset, b_vsi_uri=True, prefix: str = "") -> str:
    """
    construct the uri for the resource in the asset.
    :param asset:
    :param b_vsi_uri:
    :param prefix:
    :return:
    """

    if not asset.bucket or not asset.object_path:
        if not b_vsi_uri:
            raise FileNotFoundError("The bucket ref is not AWS or Google:\nhref : {0}".format(asset.href))
        return '/vsicurl_streaming/{}'.format(asset.href)
    elif not prefix:
        prefix = "{0}://"
        if b_vsi_uri:
            prefix = "/vsi{0}_streaming"

        if asset.cloud_platform == CloudPlatform.GCP:
            prefix = prefix.format("gs")
        elif asset.cloud_platform == CloudPlatform.AWS:
            prefix = prefix.format("s3")
        else:
            raise ValueError("The only current cloud platforms are GCP and AWS. This asset doesn't have the "
                             "'cloud_platform' field defined")

    return "{0}/{1}/{2}".format(prefix, asset.bucket, asset.object_path)


def pb_timestampfield(rel_type: FilterRelationship,
                      value: Union[datetime.datetime, datetime.date] = None,
                      start: Union[datetime.datetime, datetime.date] = None,
                      end: Union[datetime.datetime, datetime.date] = None,
                      sort_direction: SortDirection = SortDirection.NOT_SORTED,
                      tzinfo: datetime.timezone = datetime.timezone.utc) -> TimestampFilter:
    """
    Create a protobuf query filter for a timestamp or a range of timestamps. If you use a datetime.date as
    the value combined with a rel_type of EQ then you will be creating a query filter for the
    24 period of that date.
    :param rel_type: the relationship type to query more
    [here](https://geo-grpc.github.io/api/#epl.protobuf.FieldRelationship)
    :param value: time to search by using >, >=, <, <=, etc. cannot be used with start or end
    :param start: start time for between/not between query. cannot be used with value
    :param end: end time for between/not between query. cannot be used with value
    :param sort_direction: sort direction for results. Defaults to not sorting by this field
    :param tzinfo: timezone info, defaults to UTC
    :return: TimestampFilter
    """
    if rel_type in UNSUPPORTED_TIME_FILTERS:
        raise ValueError("unsupported relationship type: {}".format(rel_type.name))

    if value is not None and rel_type != FilterRelationship.EQ and rel_type != FilterRelationship.NEQ:
        if not isinstance(value, datetime.datetime):
            if rel_type == FilterRelationship.GTE or rel_type == FilterRelationship.LT:
                return TimestampFilter(value=pb_timestamp(value, tzinfo, b_force_min=True),
                                       rel_type=rel_type,
                                       sort_direction=sort_direction)
            elif rel_type == FilterRelationship.LTE or rel_type == FilterRelationship.GT:
                return TimestampFilter(value=pb_timestamp(value, tzinfo, b_force_min=False),
                                       rel_type=rel_type,
                                       sort_direction=sort_direction)
        return TimestampFilter(value=pb_timestamp(value, tzinfo), rel_type=rel_type, sort_direction=sort_direction)
    elif value is not None and not isinstance(value, datetime.datetime) and \
            (rel_type == FilterRelationship.EQ or rel_type == FilterRelationship.NEQ):
        start = datetime.datetime.combine(value, datetime.datetime.min.time(), tzinfo=tzinfo)
        end = datetime.datetime.combine(value, datetime.datetime.max.time(), tzinfo=tzinfo)
        if rel_type == FilterRelationship.EQ:
            rel_type = FilterRelationship.BETWEEN
        else:
            rel_type = FilterRelationship.NOT_BETWEEN

    return TimestampFilter(start=pb_timestamp(start, tzinfo),
                           end=pb_timestamp(end, tzinfo),
                           rel_type=rel_type,
                           sort_direction=sort_direction)


def pb_timestamp(d_utc: Union[datetime.datetime, datetime.date],
                 tzinfo: datetime.timezone = datetime.timezone.utc,
                 b_force_min=True) -> timestamp_pb2.Timestamp:
    """
    create a google.protobuf.Timestamp from a python datetime
    :param d_utc: python datetime or date
    :param tzinfo:
    :return:
    """
    ts = timestamp_pb2.Timestamp()
    ts.FromDatetime(timezoned(d_utc, tzinfo, b_force_min))
    return ts


def datetime_from_pb_timestamp(ts: timestamp_pb2.Timestamp) -> datetime:
    return datetime.datetime.utcfromtimestamp(ts.seconds + ts.nanos/1e9)


def timezoned(d_utc: Union[datetime.datetime, datetime.date],
              tzinfo: datetime.timezone = datetime.timezone.utc,
              b_force_min=True) -> datetime.datetime:
    # datetime is child to datetime.date, so if we reverse the order of this instance of we fail
    if isinstance(d_utc, datetime.datetime) and d_utc.tzinfo is None:
        # TODO add warning here:
        # print("warning, no timezone provided with datetime, so UTC is assumed")
        d_utc = datetime.datetime(d_utc.year,
                                  d_utc.month,
                                  d_utc.day,
                                  d_utc.hour,
                                  d_utc.minute,
                                  d_utc.second,
                                  d_utc.microsecond,
                                  tzinfo=tzinfo)
    elif not isinstance(d_utc, datetime.datetime):
        # print("warning, no timezone provided with date, so UTC is assumed")
        if b_force_min:
            d_utc = datetime.datetime.combine(d_utc, datetime.datetime.min.time(), tzinfo=tzinfo)
        else:
            d_utc = datetime.datetime.combine(d_utc, datetime.datetime.max.time(), tzinfo=tzinfo)
    return d_utc


# TODO https://pypi.org/project/Deprecated/
def duration(d_start: Union[datetime.datetime, datetime.date], d_end: Union[datetime.datetime, datetime.date]):
    d = duration_pb2.Duration()
    d.FromTimedelta(timezoned(d_end) - timezoned(d_start))
    return d


# TODO https://pypi.org/project/Deprecated/
def datetime_range(d_start: Union[datetime.datetime, datetime.date],
                   d_end: Union[datetime.datetime, datetime.date]) -> DatetimeRange:
    """
    for datetime range definitions for Mosaic objects.
    :param d_start: start datetime or date
    :param d_end: end datetime or date
    :return: DatetimeRange object
    """
    return DatetimeRange(start=pb_timestamp(d_start), end=pb_timestamp(d_end))


def stac_request_to_b64(req: StacRequest) -> str:
    return str(base64.b64encode(req.SerializeToString()), encoding='ascii')


def stac_request_from_b64(encoded: str) -> StacRequest:
    req = StacRequest()
    req.ParseFromString(base64.b64decode(bytes(encoded, encoding='ascii')))
    return req


def stac_item_to_b64(item: StacItem) -> str:
    return str(base64.b64encode(item.SerializeToString()), encoding='ascii')


def stac_item_from_b64(encoded: str) -> StacItem:
    item = StacItem()
    item.ParseFromString(base64.b64decode(bytes(encoded, encoding='ascii')))
    return item


def eval_float_filter(float_filter: FloatFilter, val: float) -> bool:
    rel = float_filter.rel_type

    if rel == enum.FilterRelationship.EQ:
        return val == float_filter.value
    elif rel == enum.FilterRelationship.NEQ:
        return val != float_filter.value
    elif rel == enum.FilterRelationship.BETWEEN:
        return float_filter.start < val < float_filter.end
    elif rel == enum.FilterRelationship.NOT_BETWEEN:
        return val < float_filter.start or val > float_filter.end
    elif rel == enum.FilterRelationship.GTE:
        return val >= float_filter.value
    elif rel == enum.FilterRelationship.GT:
        return val > float_filter.value
    elif rel == enum.FilterRelationship.LTE:
        return val <= float_filter.value
    elif rel == enum.FilterRelationship.LT:
        return val < float_filter.value
    else:
        raise ValueError(f"not currently evaluating float filters of type: {query.FilterRelationship.Name(rel)}")



import json

from base64 import b64decode
from datetime import datetime, timezone

from epl.protobuf.v1 import stac_pb2

from nsl.stac.destinations import BaseDestination, DestinationDecoder
from nsl.stac.utils import stac_request_from_b64


class Subscription:
    id: str
    nsl_id: str
    is_active: bool
    created_at: datetime
    stac_request: stac_pb2.StacRequest
    destination: BaseDestination

    def __init__(self, response_dict: dict):
        self.id = response_dict['sub_id']
        self.nsl_id = response_dict['nsl_id']
        self.is_active = response_dict['is_active']
        self.created_at = datetime.utcfromtimestamp(response_dict['created_at']).replace(tzinfo=timezone.utc)
        self.stac_request = stac_request_from_b64(response_dict['stac_request'])
        self.destination = json.loads(response_dict['destination_json'], cls=DestinationDecoder)



from pathlib import Path
from typing import Union

from google.cloud.storage import Blob, Bucket

from epl.protobuf.v1 import stac_pb2
from nsl.stac import gcs_storage_client
from nsl.stac.enum import AssetType
from nsl.stac.destinations.base import BaseDestination


class GCPDestination(BaseDestination):
    # TODO: GKE access to a bucket w/in the same region is Free
    # TODO: "transfer" between buckets w/in the same region is Free
    type = 'gcp'

    save_directory: Path
    bucket: str
    region: str

    def __init__(self,
                 bucket: str,
                 region: str,
                 asset_type: AssetType = AssetType.GEOTIFF,
                 save_directory: Union[Path, str] = '/'):
        super().__init__(asset_type=asset_type)
        self.save_directory = Path(save_directory)
        self.bucket = bucket
        self.region = region

    def deliver(self, nsl_id: str, sub_id: str, stac_item: stac_pb2.StacItem):
        try:
            # TODO: in the future, check a local cache for this asset to avoid multiple downloads
            self.target_blob(stac_item)\
                .upload_from_file(self.src_blob(stac_item).open('rb'), client=gcs_storage_client.client)
            return None
        except BaseException as err:
            print(f'ERROR: failed to transfer asset {stac_item.id}:\n{err}')
            raise err

    def __json__(self) -> dict:
        return dict(**super().__json__(),
                    bucket=self.bucket,
                    region=self.region,
                    save_directory=str(self.save_directory))

    def target_blob(self, stac_item: stac_pb2.StacItem) -> Blob:
        return Blob(name=self.blob_path(stac_item, self.save_directory),
                    bucket=Bucket(client=gcs_storage_client.client, name=self.bucket))



import pathlib
import re
import typing
import uuid

from copy import deepcopy
from datetime import date, datetime, timezone
from typing import BinaryIO, Dict, IO, Iterator, List, Optional, Set, Tuple, Union

import boto3
import botocore.exceptions
from google.protobuf.any_pb2 import Any
from google.protobuf.wrappers_pb2 import FloatValue
from shapely.geometry.base import BaseGeometry
from shapely.geometry import Polygon

from nsl.stac import enum, utils, stac_service as stac_singleton, \
    StacItem, StacRequest, Collection, CollectionRequest, View, ViewRequest, Mosaic, MosaicRequest, Eo, EoRequest, \
    Extent, Interval, Provider, ProjectionData, GeometryData, EnvelopeData, \
    Asset, FloatFilter, StringFilter, TimestampFilter
from nsl.stac.client import NSLClient
from nsl.stac.destinations import BaseDestination
from nsl.stac.subscription import Subscription


class ProviderRole:
    LICENSOR = 'licensor'
    PRODUCER = 'producer'
    PROCESSOR = 'processor'
    HOST = 'host'

    @staticmethod
    def all() -> List[str]:
        return [ProviderRole.LICENSOR, ProviderRole.PRODUCER, ProviderRole.PROCESSOR, ProviderRole.HOST]

    @staticmethod
    def nsl() -> Provider:
        provider = Provider()
        provider.name = "Near Space Labs"
        provider.description = "https://nearspacelabs.com"
        provider.roles[:] = ProviderRole.all()
        provider.url = "https://nearspacelabs.com"
        return provider

    @staticmethod
    def nsl_with_roles(roles: List[str]) -> Provider:
        provider = Provider()
        provider.CopyFrom(Provider.nsl())
        assert all(role in ProviderRole.all() for role in roles)
        provider.roles[:] = roles
        return provider


def _from_protobuf(geometry_data: GeometryData) -> BaseGeometry:
    if len(geometry_data.wkt) > 0:
        from shapely.wkt import loads as loads_wkt
        return loads_wkt(geometry_data.wkt)
    elif len(geometry_data.wkb) > 0:
        from shapely.wkb import loads as loads_wkb
        return loads_wkb(geometry_data.wkb)
    else:
        raise ValueError("no geometry data")


def _from_envelope_data(envelope_data: EnvelopeData) -> BaseGeometry:
    return Polygon.from_bounds(xmin=envelope_data.xmin,
                               ymin=envelope_data.ymin,
                               xmax=envelope_data.xmax,
                               ymax=envelope_data.ymax)


def _to_protobuf(geometry: BaseGeometry, proj: ProjectionData = None):
    if proj is None:
        print("warning, no projection data set. assuming WGS84")
        proj = ProjectionData(epsg=4326)
    return GeometryData(wkb=geometry.wkb, proj=proj)


def _to_envelope_data(geometry: BaseGeometry, proj: ProjectionData = None):
    if proj is None:
        print("warning, no projection data set. assuming WGS84")
        proj = ProjectionData(epsg=4326)
    return EnvelopeData(xmin=geometry.bounds[0],
                        ymin=geometry.bounds[1],
                        xmax=geometry.bounds[2],
                        ymax=geometry.bounds[3],
                        proj=proj)


def _set_properties(stac_data, properties, type_url_prefix):
    """
     pack properties and then set the properties member value to the input.
     :param stac_data:
     :param properties:
     :param type_url_prefix:
     :return:
     """
    if properties is None:
        return

    # pack the properties into an Any field
    packed_properties = Any()
    packed_properties.Pack(properties,
                           type_url_prefix=type_url_prefix + properties.DESCRIPTOR.full_name)

    # overwrite the previous properties field with this updated version
    stac_data.properties.CopyFrom(packed_properties)
    properties = properties
    return stac_data, properties


def _check_assets_exist(stac_item: StacItem, b_raise=True) -> List[str]:
    results = []
    for asset_key in stac_item.assets:
        asset = stac_item.assets[asset_key]
        b_file_exists = _check_asset_exists(asset)

        if not b_file_exists and b_raise:
            raise ValueError("get_blob_metadata returns false for asset key {}".format(asset_key))
        results.append(asset_key)
    return results


def _check_asset_exists(asset: Asset) -> bool:
    if asset.cloud_platform == enum.CloudPlatform.GCP:
        return utils.get_blob_metadata(bucket=asset.bucket, blob_name=asset.object_path) is not None
    elif asset.cloud_platform == enum.CloudPlatform.AWS:
        return _check_aws_asset_exists(asset)
    else:
        raise ValueError("cloud platform {0} of asset {1} not supported"
                         .format(enum.CloudPlatform(asset.cloud_platform).name, asset))


def _check_aws_asset_exists(asset: Asset) -> bool:
    s3 = boto3.client('s3')

    try:
        s3.head_object(Bucket=asset.bucket, Key=asset.object_path, RequestPayer='requester')
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == "404":
            return False
        else:
            raise e
    return True


class AssetWrap(object):
    def __init__(self,
                 asset: Asset = None,
                 bucket: str = None,
                 object_path: str = None,
                 asset_type: enum.AssetType = enum.AssetType.UNKNOWN_ASSET,
                 eo_bands: enum.Band = enum.Band.UNKNOWN_BAND,
                 href="",
                 cloud_platform: enum.CloudPlatform = enum.CloudPlatform.UNKNOWN_CLOUD_PLATFORM,
                 bucket_manager: str = "",
                 bucket_region: str = "",
                 key_suffix: str = "",
                 asset_key: str = ""):
        self._asset_key = asset_key
        if asset is None:
            asset = Asset(href=href,
                          eo_bands=eo_bands,
                          asset_type=asset_type,
                          cloud_platform=cloud_platform,
                          bucket_manager=bucket_manager,
                          bucket_region=bucket_region,
                          bucket=bucket,
                          object_path=object_path)

        self._asset = asset

        if self._asset.bucket is None or self._asset.object_path is None:
            raise ValueError("bucket and object path must be set in valid asset")

        self._ext = pathlib.Path(self._asset.object_path).suffix

        b_thumbnail_png = self._asset.asset_type == enum.AssetType.THUMBNAIL and self._ext == '.png'
        _, href_type = self._asset_type_details(asset_type=self._asset.asset_type, b_thumbnail_png=b_thumbnail_png)
        self._asset.type = href_type
        self._type = href_type
        self._key_suffix = key_suffix

    def __eq__(self, other):
        if not isinstance(other, AssetWrap):
            return False

        if self.asset_key != other.asset_key:
            return False

        return self.equals_pb(other._asset)

    def __str__(self):
        return str("{0}extension: {1}\nasset_key: {2}".format(self._asset, self._ext, self.asset_key))

    def __copy__(self):
        pass

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == '_asset':
                value = Asset()
                value.CopyFrom(v)
                setattr(result, k, value)
            else:
                setattr(result, k, deepcopy(v))
        return result

    def _asset_key_prefix(self) -> str:
        if self.eo_bands == enum.Band.UNKNOWN_BAND:
            return "{0}_{1}".format(self.asset_type.name, self.cloud_platform.name)
        return "{0}_{1}_{2}".format(self.asset_type.name, self.eo_bands.name, self.cloud_platform.name)

    @property
    def asset(self) -> Asset:
        return self._asset

    @property
    def asset_key(self) -> str:
        if self._asset_key:
            return self._asset_key

        asset_key = self._asset_key_prefix()
        if self._key_suffix:
            asset_key += "_{}".format(self._key_suffix)
        return asset_key

    @property
    def asset_key_suffix(self) -> str:
        return self._key_suffix

    @asset_key_suffix.setter
    def asset_key_suffix(self, value: str):
        self._key_suffix = value

    @property
    def asset_type(self) -> enum.AssetType:
        """
type of asset
        :return:
        """
        return enum.AssetType(self._asset.asset_type)

    @property
    def bucket(self) -> str:
        """
bucket where data stored. may be private
        :return:
        """
        return self._asset.bucket

    @property
    def bucket_manager(self) -> str:
        return self._asset.bucket_manager

    @property
    def bucket_region(self) -> str:
        return self._asset.bucket_region

    @property
    def cloud_platform(self) -> enum.CloudPlatform:
        """
cloud platform where data stored. Google Cloud, AWS and Azure are current options (or unknown)
        :return:
        """
        return enum.CloudPlatform(self._asset.cloud_platform)

    @cloud_platform.setter
    def cloud_platform(self, value: enum.CloudPlatform):
        self._asset.cloud_platform = value

    @property
    def eo_bands(self) -> enum.Band:
        """
electro optical bands included in data. if data is not electro optical, then this is set to Unknown
        :return:
        """
        return enum.Band(self._asset.eo_bands)

    @property
    def ext(self) -> str:
        return self._ext

    @property
    def href(self) -> str:
        """
the href for downloading data
        :return:
        """
        return self._asset.href

    @href.setter
    def href(self, value: str):
        self._asset.href = value

    @property
    def object_path(self) -> str:
        """
the object path to use with the bucket if access is available
        :return:
        """
        return self._asset.object_path

    @object_path.setter
    def object_path(self, value: str):
        self._asset.object_path = value

    @property
    def type(self) -> str:
        return self._type

    def equals_pb(self, other: Asset):
        """
does the AssetWrap equal a protobuf Asset
        :param other:
        :return:
        """
        return self._asset.SerializeToString() == other.SerializeToString()

    def exists(self) -> bool:
        return _check_asset_exists(self._asset)

    def download(self,
                 from_bucket: bool = False,
                 file_obj: IO[Union[Union[str, bytes], Any]] = None,
                 save_filename: str = '',
                 save_directory: str = '',
                 requester_pays: bool = False,
                 nsl_id: str = None,
                 profile_name: str = None) -> str:
        return utils.download_asset(asset=self._asset,
                                    from_bucket=from_bucket,
                                    file_obj=file_obj,
                                    save_filename=save_filename,
                                    save_directory=save_directory,
                                    requester_pays=requester_pays,
                                    nsl_id=nsl_id,
                                    profile_name=profile_name)

    def matches_details(self,
                        asset_key: str = None,
                        asset_type: enum.AssetType = enum.AssetType.UNKNOWN_ASSET,
                        cloud_platform: enum.CloudPlatform = enum.CloudPlatform.UNKNOWN_CLOUD_PLATFORM,
                        eo_bands: enum.Band = enum.Band.UNKNOWN_BAND,
                        asset_regex: Dict = None,
                        b_relaxed_types: bool = False) -> bool:
        if asset_key is not None and asset_key != self.asset_key:
            return False

        b_asset_type_match = self._asset_types_match(desired_type=asset_type,
                                                     asset_type=self.asset_type,
                                                     b_relaxed_types=b_relaxed_types)

        asset_key = self.asset_key

        if (eo_bands is not None and eo_bands != enum.Band.UNKNOWN_BAND) and self.eo_bands != eo_bands:
            return False
        if (cloud_platform is not None and cloud_platform != enum.CloudPlatform.UNKNOWN_CLOUD_PLATFORM) and \
                self.cloud_platform != cloud_platform:
            return False
        if (asset_type is not None and asset_type != enum.AssetType.UNKNOWN_ASSET) and not b_asset_type_match:
            return False

        if asset_regex is not None and len(asset_regex) > 0:
            for key, regex_value in asset_regex.items():
                if key == 'asset_key':
                    if not re.match(regex_value, asset_key):
                        return False
                else:
                    if not hasattr(self, key):
                        raise AttributeError("no key {0} in asset {1}".format(key, self))
                    elif not re.match(regex_value, getattr(self, key)):
                        return False

        return True

    @staticmethod
    def _asset_types_match(desired_type: enum.AssetType, asset_type: enum.AssetType,
                           b_relaxed_types: bool = False) -> bool:
        if not b_relaxed_types:
            return desired_type == asset_type
        elif desired_type == enum.AssetType.TIFF:
            return asset_type == desired_type or \
                   asset_type == enum.AssetType.GEOTIFF or \
                   asset_type == enum.AssetType.CO_GEOTIFF
        elif desired_type == enum.AssetType.GEOTIFF:
            return asset_type == desired_type or asset_type == enum.AssetType.CO_GEOTIFF
        return asset_type == desired_type

    @staticmethod
    def _asset_type_details(asset_type: enum.AssetType, b_thumbnail_png: bool = True) -> Tuple[str, str]:
        """
for asset type and bool, get the extension and href type
        :param asset_type:
        :param b_thumbnail_png:
        :return: str extension, str href type
        """
        # TODO finish asset_type map
        ext = "tif"
        if asset_type == enum.AssetType.TIFF:
            href_type = "image/tiff"
        elif asset_type == enum.AssetType.GEOTIFF:
            href_type = "image/vnd.stac.geotiff"
        elif asset_type == enum.AssetType.CO_GEOTIFF:
            href_type = "image/vnd.stac.geotiff; cloud-optimized=true"
        elif (asset_type == enum.AssetType.THUMBNAIL and b_thumbnail_png) or asset_type == enum.AssetType.PNG:
            href_type = "image/png"
            ext = "png"
        elif (asset_type == enum.AssetType.THUMBNAIL and not b_thumbnail_png) or asset_type == enum.AssetType.JPEG:
            href_type = "image/jpeg"
            ext = "jpg"
        elif asset_type == enum.AssetType.JPEG_2000:
            href_type = "image/jp2"
            ext = "jp2"
        elif asset_type == enum.AssetType.MRF_XML:
            href_type = "application/xml"
            ext = "xml"
        else:
            href_type = "application/octet-stream"
            ext = "bin"

        return ext, href_type


class _BaseWrap:
    def __init__(self,
                 stac_data,
                 properties_func,
                 type_url_prefix="nearspacelabs.com/proto/",
                 correlation_id: str = None):
        """
        Whether it's a stac_request or a stac_item, allow for the repack_properties method to work
        :param stac_data:
        :param properties_func:
        """
        self._stac_data = stac_data
        self.properties = None
        self._properties_func = properties_func
        self._type_url_prefix = type_url_prefix
        self._correlation_id = correlation_id if correlation_id is not None else str(uuid.uuid4())

        if stac_data is not None and properties_func is not None and stac_data.HasField("properties"):
            self.properties = properties_func()
            self._stac_data.properties.Unpack(self.properties)
        elif properties_func is not None:
            self.properties = properties_func()
            self._set_properties(self.properties)

    def __str__(self):
        return str(self._stac_data)

    def _set_properties(self, properties):
        self._stac_data, self.properties = _set_properties(self._stac_data, properties, self._type_url_prefix)

    def _get_field(self, metadata_key: str, key: str) -> Optional[typing.Any]:
        if self.properties.HasField(metadata_key):
            return getattr(getattr(self.properties, metadata_key), key)
        return None

    def _get_wrapped_field(self, metadata_key: str, key: str) -> Optional[typing.Any]:
        if self.properties.HasField(metadata_key):
            return getattr(getattr(getattr(self.properties, metadata_key), key), "value")
        return None

    def _set_internal_sub_object(self, metadata_key: str):
        pass

    def _set_field(self, metadata_key: str, key: str, value):
        self._set_internal_sub_object(metadata_key)
        setattr(getattr(self.properties, metadata_key), key, value)

    def _set_obj(self, metadata_key: str, key: str, value):
        self._set_internal_sub_object(metadata_key)
        getattr(getattr(self.properties, metadata_key), key).CopyFrom(value)

    def _set_nested_obj(self, metadata_key: str, object_key: str, value_key: str, value):
        self._set_internal_sub_object(metadata_key)
        getattr(getattr(getattr(self.properties, metadata_key), object_key), value_key).CopyFrom(value)

    def _set_nested_field(self, metadata_key: str, object_key: str, value_key: str, value):
        setattr(getattr(getattr(self.properties, metadata_key), object_key), value_key, value)

    def _get_nested_field(self, metadata_key: str, object_key: str, value_key: str) -> Optional[typing.Any]:
        if self.properties.HasField(metadata_key):
            return getattr(getattr(getattr(self.properties, metadata_key), object_key), value_key)
        return None

    def _get_nested_wrapped_field(self, metadata_key: str, object_key: str, value_key: str) -> Optional[typing.Any]:
        if self.properties.HasField(metadata_key):
            return getattr(getattr(getattr(getattr(self.properties, metadata_key), object_key), value_key), "value")
        return None

    @property
    def correlation_id(self) -> str:
        return self._correlation_id

    @correlation_id.setter
    def correlation_id(self, correlation_id: str):
        self._correlation_id = correlation_id


class StacItemWrap(_BaseWrap):
    """
Wrapper for StacItem protobuf
    """

    def __eq__(self, other):
        if not isinstance(other, StacItemWrap):
            return False

        return self.equals_pb(other.stac_item)

    def __init__(self, stac_item: StacItem = None, properties_constructor=None, correlation_id: str = None):
        self._assets = {}
        if stac_item is None:
            stac_data = StacItem()
        else:
            stac_data = StacItem()
            stac_data.CopyFrom(stac_item)

            for asset_key in stac_data.assets:
                self._assets[asset_key] = AssetWrap(stac_data.assets[asset_key], asset_key=asset_key)

        super().__init__(stac_data=stac_data, properties_func=properties_constructor, correlation_id=correlation_id)
        if self.created is None:
            self.created = datetime.now(tz=timezone.utc)

    def __copy__(self):
        pass

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == '_stac_data':
                value = StacItem()
                value.CopyFrom(v)
                setattr(result, k, value)
            else:
                setattr(result, k, deepcopy(v))
        return result

    @property
    def bbox(self) -> EnvelopeData:
        """
bounding box of data. In form of EnvelopeData
        :return:
        """
        return self.stac_item.bbox

    @property
    def cloud_cover(self) -> Optional[float]:
        """
get cloud cover value
        :return: float or None
        """
        if self.stac_item.HasField("eo") and self.stac_item.eo.HasField("cloud_cover"):
            return self.stac_item.eo.cloud_cover.value
        return None

    @cloud_cover.setter
    def cloud_cover(self, value: float):
        if not self.stac_item.HasField("eo"):
            self.stac_item.eo.CopyFrom(Eo(cloud_cover=FloatValue(value=value)))
        else:
            self.stac_item.eo.cloud_cover.CopyFrom(FloatValue(value=value))

    @property
    def collection(self) -> str:
        """
the collection id for the stac item
        :return:
        """
        return self.stac_item.collection

    @collection.setter
    def collection(self, value: str):
        self.stac_item.collection = value

    @property
    def constellation(self) -> enum.Constellation:
        """
the enum describing the constellation
        :return:
        """
        return enum.Constellation(self.stac_item.constellation_enum)

    @constellation.setter
    def constellation(self, value: enum.Constellation):
        self.stac_item.constellation_enum = value
        self.stac_item.constellation = self.constellation.name

    @property
    def created(self) -> Optional[datetime]:
        if self.stac_item.HasField("created"):
            return datetime.fromtimestamp(self.stac_item.created.seconds, tz=timezone.utc)
        else:
            return None

    @created.setter
    def created(self, value: Union[datetime, date]):
        self.stac_item.created.CopyFrom(utils.pb_timestamp(d_utc=value))

    @property
    def end_observed(self) -> Optional[datetime]:
        return self.stac_item.end_observation

    @end_observed.setter
    def end_observed(self, value: Union[datetime, date]):
        self.stac_item.end_observation.CopyFrom(utils.pb_timestamp(d_utc=value))
        self.stac_item.end_datetime.CopyFrom(utils.pb_timestamp(d_utc=value))

    @property
    def feature(self):
        """
        geojson feature with geometry being only aspect defined
        :return:
        """
        return {
            'type': 'Feature',
            'geometry': self.geometry.__geo_interface__,
            'id': self.id,
            'collection': self.collection,
            'properties': self._feature_properties(),
            'assets': self._feature_assets()
        }

    @staticmethod
    def _append_prop(props: Dict, prop_name: str, prop_value):
        if prop_value is None:
            return props
        if props is None:
            props = {}
        props[prop_name] = prop_value
        return props

    def _feature_properties(self) -> Dict:
        props = {}
        props = StacItemWrap._append_prop(props, 'datetime', self.observed.replace(microsecond=0).isoformat())
        props = StacItemWrap._append_prop(props, 'observed', self.observed.replace(microsecond=0).isoformat())
        props = StacItemWrap._append_prop(props, 'mission', self.mission.name)
        props = StacItemWrap._append_prop(props, 'platform', self.platform.name)
        props = StacItemWrap._append_prop(props, 'instrument', self.instrument.name)
        props = StacItemWrap._append_prop(props, 'gsd', self.gsd)
        props = StacItemWrap._append_prop(props, 'eo:cloud_cover', self.cloud_cover)
        props = StacItemWrap._append_prop(props, 'view:off_nadir', self.off_nadir)
        return props

    def _feature_assets(self) -> Dict:
        feature_assets = {}
        for asset_wrap in self.get_assets():
            feature_assets[asset_wrap.asset_key] = {
                'href': asset_wrap.href,
                'type': asset_wrap.type,
            }
        return feature_assets

    @property
    def geometry(self) -> BaseGeometry:
        if self.stac_item.HasField("geometry"):
            return _from_protobuf(self.stac_item.geometry)
        elif self.stac_item.HasField("bbox"):
            return _from_envelope_data(self.stac_item.bbox)

    @geometry.setter
    def geometry(self, value: BaseGeometry):
        if isinstance(value, BaseGeometry):
            self.stac_item.geometry.CopyFrom(_to_protobuf(value))
        else:
            # try epl.geometry
            self.stac_item.geometry.CopyFrom(value.geometry_data)
        self.stac_item.bbox.CopyFrom(_to_envelope_data(value))

    @property
    def geometry_proj(self) -> ProjectionData:
        if self.stac_item.HasField("geometry"):
            return self.stac_item.geometry.proj
        elif self.stac_item.HasField("bbox"):
            return self.stac_item.bbox.proj
        return None

    @geometry_proj.setter
    def geometry_proj(self, value: ProjectionData):
        if self.stac_item.HasField("geometry"):
            self.stac_item.geometry.proj.CopyFrom(value)
            return
        raise ValueError("can't set geometry projection without first setting geometry")

    @property
    def gsd(self) -> Optional[float]:
        """
        get cloud cover value
        :return: float or None
        """
        if self.stac_item.HasField("gsd"):
            return self.stac_item.gsd.value
        return None

    @gsd.setter
    def gsd(self, value: float):
        self.stac_item.gsd.CopyFrom(FloatValue(value=value))

    @property
    def id(self) -> str:
        return self.stac_item.id

    @id.setter
    def id(self, value: str):
        self.stac_item.id = value

    @property
    def instrument(self) -> enum.Instrument:
        return enum.Instrument(self.stac_item.instrument_enum)

    @instrument.setter
    def instrument(self, value: enum.Instrument):
        self.stac_item.instrument_enum = value
        self.stac_item.instrument = self.instrument.name

    @property
    def mission(self) -> enum.Mission:
        return enum.Mission(self.stac_item.mission_enum)

    @mission.setter
    def mission(self, value: enum.Mission):
        self.stac_item.mission_enum = value
        self.stac_item.mission = self.mission.name

    @property
    def mosaic_name(self) -> Optional[str]:
        if self.stac_item.HasField("mosaic"):
            return self.stac_item.mosaic.name
        return None

    @mosaic_name.setter
    def mosaic_name(self, name: str):
        if not self.stac_item.HasField("mosaic"):
            self.stac_item.mosaic.CopyFrom(Mosaic(name=name))
        else:
            self.stac_item.mosaic.name = name

    @property
    def mosaic_quad_key(self) -> Optional[str]:
        """
        If the STAC item is a quad from a mosaic, then it has a quad key that defines the boundaries of the quad. The
        quad tree definition is assumed to be the convention defined by Google Maps, based off of there Pseudo-Web
        Mercator projection.

        An example quad key is '02313012030231'. Quads use 2-bit tile interleaved addresses. The first character
        defines the largest quadrant (in this case 0 is upper left), the next character ('2') is the upper right
        quadrant of that first quadrant, the 3rd character ('3') is the lower left quadrant of the previous quadrant
        and so on.

        For more details on the quad tree tiling for maps use `openstreetmaps docs
        <https://wiki.openstreetmap.org/wiki/QuadTiles#Quadtile_implementation>`
        :return:
        """
        if self.stac_item.HasField("mosaic"):
            return self.stac_item.mosaic.quad_key
        return None

    @mosaic_quad_key.setter
    def mosaic_quad_key(self, quad_key: str):
        if not self.stac_item.HasField("mosaic"):
            self.stac_item.mosaic.CopyFrom(Mosaic(quad_key=quad_key))
        else:
            self.stac_item.mosaic.quad_key = quad_key

    @property
    def observed(self) -> Optional[datetime]:
        if self.stac_item.HasField("datetime"):
            return datetime.fromtimestamp(self.stac_item.datetime.seconds, tz=timezone.utc)
        elif self.stac_item.HasField("observed"):
            return datetime.fromtimestamp(self.stac_item.observed.seconds, tz=timezone.utc)
        else:
            return None

    @observed.setter
    def observed(self, value: Union[datetime, date]):
        self.stac_item.datetime.CopyFrom(utils.pb_timestamp(d_utc=value))
        self.stac_item.observed.CopyFrom(utils.pb_timestamp(d_utc=value))

    @property
    def off_nadir(self) -> Optional[float]:
        """
        get cloud cover value
        :return: float or None
        """
        if self.stac_item.HasField("view") and self.stac_item.view.HasField("off_nadir"):
            return self.stac_item.view.off_nadir.value
        return None

    @off_nadir.setter
    def off_nadir(self, value: float):
        if not self.stac_item.HasField("view"):
            self.stac_item.view.CopyFrom(View(off_nadir=FloatValue(value=value)))
        else:
            self.stac_item.view.off_nadir.CopyFrom(FloatValue(value=value))

    @property
    def platform(self) -> enum.Platform:
        return enum.Platform(self.stac_item.platform_enum)

    @platform.setter
    def platform(self, value: enum.Platform):
        self.stac_item.platform_enum = value
        self.stac_item.platform = self.platform.name

    @property
    def provenance_ids(self) -> List[str]:
        """
        The stac_ids that went into creating the current mosaic. They are in the array in the order which they were
        used in the mosaic
        :return:
        """
        return self.stac_item.mosaic.provenance_ids

    @property
    def proj(self) -> ProjectionData:
        """
        The projection for all assets of this STAC item. If an Asset has its own proj definition,
        then that supersedes this projection definition.
        :return: projection information
        """
        return self.stac_item.proj

    @proj.setter
    def proj(self, value: ProjectionData):
        self.stac_item.proj.CopyFrom(value)

    @property
    def stac_item(self) -> StacItem:
        """
        get stac_item
        :return: StacItem
        """
        return self._stac_data

    @property
    def updated(self) -> Optional[datetime]:
        if self.stac_item.HasField("updated"):
            return datetime.fromtimestamp(self.stac_item.updated.seconds, tz=timezone.utc)
        else:
            return None

    @updated.setter
    def updated(self, value: Union[datetime, date]):
        self.stac_item.updated.CopyFrom(utils.pb_timestamp(d_utc=value))

    def download_asset(self,
                       asset_key: str = "",
                       asset_type: enum.AssetType = enum.AssetType.UNKNOWN_ASSET,
                       cloud_platform: enum.CloudPlatform = enum.CloudPlatform.GCP,
                       eo_bands: enum.Band = enum.Band.UNKNOWN_BAND,
                       asset_key_regex: str = None,
                       from_bucket: bool = False,
                       file_obj: BinaryIO = None,
                       save_filename: str = "",
                       save_directory: str = "",
                       nsl_id: str = None,
                       profile_name: str = None) -> str:
        asset_wrap = self.get_asset(asset_key=asset_key,
                                    asset_type=asset_type,
                                    cloud_platform=cloud_platform,
                                    eo_bands=eo_bands,
                                    asset_regex=asset_key_regex)

        return asset_wrap.download(from_bucket=from_bucket,
                                   file_obj=file_obj,
                                   save_filename=save_filename,
                                   save_directory=save_directory,
                                   nsl_id=nsl_id,
                                   profile_name=profile_name)

    def equals_pb(self, other: StacItem):
        """
        does the StacItemWrap equal a protobuf StacItem
        :param other:
        :return:
        """
        return self.stac_item.SerializeToString() == other.SerializeToString()

    def has_asset(self,
                  asset_key: str = None,
                  asset_type: enum.AssetType = enum.AssetType.UNKNOWN_ASSET,
                  cloud_platform: enum.CloudPlatform = enum.CloudPlatform.UNKNOWN_CLOUD_PLATFORM,
                  eo_bands: enum.Band = enum.Band.UNKNOWN_BAND,
                  asset_regex: Dict = None,
                  b_relaxed_types: bool = False) -> bool:
        if asset_key is not None and asset_key in self._assets:
            return True
        elif asset_key is not None and asset_key and asset_key not in self._assets:
            return False

        for asset in self._assets.values():
            if asset.matches_details(asset_key, asset_type, cloud_platform, eo_bands, asset_regex, b_relaxed_types):
                return True
        return False

    def get_assets(self,
                   asset_key: str = None,
                   asset_type: enum.AssetType = enum.AssetType.UNKNOWN_ASSET,
                   cloud_platform: enum.CloudPlatform = enum.CloudPlatform.UNKNOWN_CLOUD_PLATFORM,
                   eo_bands: enum.Band = enum.Band.UNKNOWN_BAND,
                   asset_regex: Dict = None,
                   b_relaxed_types: bool = False) -> List[AssetWrap]:
        if asset_key is not None and asset_key in self._assets:
            return [self._assets[asset_key]]
        elif asset_key is not None and asset_key and asset_key not in self._assets:
            raise ValueError("asset_key {} not found".format(asset_key))

        results = []
        for asset_key in self._assets:
            current = self._assets[asset_key]
            if not current.matches_details(asset_key, asset_type, cloud_platform, eo_bands, asset_regex, b_relaxed_types):
                continue

            # check that asset hasn't changed between protobuf and asset_map
            pb_asset = self.stac_item.assets[asset_key]
            if not current.equals_pb(pb_asset):
                raise ValueError("corrupted protobuf. Asset and AssetWrap have differing underlying protobuf")

            results.append(current)
        return results

    def get_asset(self,
                  asset_key: str = None,
                  asset_type: enum.AssetType = enum.AssetType.UNKNOWN_ASSET,
                  cloud_platform: enum.CloudPlatform = enum.CloudPlatform.UNKNOWN_CLOUD_PLATFORM,
                  eo_bands: Eo.Band = Eo.UNKNOWN_BAND,
                  asset_regex: Dict = None,
                  b_relaxed_types: bool = False) -> Optional[AssetWrap]:
        results = self.get_assets(asset_key, asset_type, cloud_platform, eo_bands, asset_regex, b_relaxed_types)
        if len(results) > 1:
            raise ValueError("must be more specific in selecting your asset. if all enums are used, try using "
                             "asset_key_regex")
        elif len(results) == 1:
            return results[0]
        return None

    def check_assets_exist(self, b_raise) -> List[str]:
        return _check_assets_exist(self.stac_item, b_raise=b_raise)

    def append_mosaic_provenance(self, provenance_stac_id: str):
        """
        Add a provenance stac_id to a mosaic. The 0th item is the first item added to the mosaic, and is therefore the
        bottom-most image in the mosaic. The -1 item is the last item added to the mosaic; ie the topmost image in the
        mosaic.
        :param provenance_stac_id:
        """
        if not self.stac_item.HasField("mosaic"):
            self.stac_item.mosaic.CopyFrom(Mosaic(provenance_ids=[provenance_stac_id]))
        else:
            self.stac_item.mosaic.provenance_ids.append(provenance_stac_id)

    def set_mosaic_datetime_range(self, d_start, d_end):
        datetime_range = utils.datetime_range(d_start=d_start, d_end=d_end)
        if not self.stac_item.HasField("mosaic"):
            self.stac_item.mosaic.CopyFrom(Mosaic(observation_range=datetime_range))
        else:
            self.stac_item.mosaic.observation_range.CopyFrom(datetime_range)


class StacRequestWrap(_BaseWrap):
    def __init__(self,
                 stac_request: StacRequest = None,
                 properties_constructor=None,
                 id: str = "",
                 correlation_id: str = None):
        if stac_request is None:
            stac_request = StacRequest(id=id)

        super().__init__(stac_data=stac_request, properties_func=properties_constructor, correlation_id=correlation_id)

    @property
    def bbox(self) -> EnvelopeData:
        if self.stac_request.HasField("bbox"):
            return self.stac_request.bbox
        elif self.stac_request.HasField("intersects"):
            # TODO mono-759, this seems messy
            return _to_envelope_data(_from_protobuf(self.intersects))
        return None

    @bbox.setter
    def bbox(self, value: EnvelopeData):
        # this tests the spatial reference (it would be better to have a dedicated method)
        self.stac_request.bbox.CopyFrom(value)
        self.stac_request.ClearField("intersects")
        # TODO mono-759, this seems messy
        self.stac_request.intersects.CopyFrom(_to_protobuf(_from_envelope_data(value), proj=value.proj))

    @property
    def collection(self) -> str:
        return self.stac_request.collection

    @collection.setter
    def collection(self, value: str):
        self.stac_request.collection = value

    @property
    def constellation(self) -> enum.Constellation:
        return enum.Constellation(self.stac_request.constellation_enum)

    @constellation.setter
    def constellation(self, value: enum.Constellation):
        self.stac_request.constellation_enum = value

    @property
    def id(self) -> str:
        return self.stac_request.id

    @id.setter
    def id(self, value: str):
        self.stac_request.id = value
        if self.stac_request.HasField("id_complex"):
            self.stac_request.ClearField("id_complex")

    @property
    def instrument(self) -> enum.Instrument:
        return enum.Instrument(self.stac_request.instrument_enum)

    @instrument.setter
    def instrument(self, value: enum.Instrument):
        self.stac_request.instrument_enum = value

    @property
    def intersects(self) -> Optional[BaseGeometry]:
        if self.stac_request.HasField("intersects"):
            return _from_protobuf(self.stac_request.intersects)
        elif self.stac_request.HasField("bbox"):
            return _from_envelope_data(self.bbox)
        return None

    @intersects.setter
    def intersects(self, geometry: BaseGeometry, proj: ProjectionData = None):
        if isinstance(geometry, BaseGeometry):
            if proj is None:
                print("warning, no projection data set. assuming WGS84")
                proj = ProjectionData(epsg=4326)
            g = GeometryData(wkb=geometry.wkb, proj=proj)
            self.stac_request.intersects.CopyFrom(g)
        else:
            # assume epl.geometry
            self.stac_request.intersects.CopyFrom(geometry.geometry_data)
        self.stac_request.ClearField("bbox")

    @property
    def intersects_proj(self) -> ProjectionData:
        if self.stac_request.HasField("intersects"):
            return self.stac_request.intersects.proj
        elif self.stac_request.HasField("bbox"):
            return self.stac_request.bbox.proj
        return None

    @intersects_proj.setter
    def intersects_proj(self, value: ProjectionData):
        if self.stac_request.HasField("geometry"):
            self.stac_request.intersects.proj.CopyFrom(value)
            return
        raise ValueError("can't set intersects projection without first setting geometry")

    @property
    def limit(self) -> int:
        return self.stac_request.limit

    @limit.setter
    def limit(self, value: int):
        self.stac_request.limit = value

    @property
    def mission(self) -> enum.Mission:
        return enum.Mission(self.stac_request.mission_enum)

    @mission.setter
    def mission(self, value: enum.Mission):
        self.stac_request.mission_enum = value

    @property
    def mosaic_name(self) -> Optional[str]:
        if self.stac_request.HasField("mosaic"):
            return self.stac_request.mosaic.name
        return None

    @mosaic_name.setter
    def mosaic_name(self, value: str):
        if not self.stac_request.HasField("mosaic"):
            self.stac_request.mosaic.CopyFrom(MosaicRequest(name=value))
        else:
            self.stac_request.mosaic.name = value

    @property
    def mosaic_quad_key(self) -> Optional[str]:
        """
Overview of :func:`~StacItemWrap.mosaic_quad_key`

The quad_key to search for mosaic quad STAC items by. If a quad STAC item exists with the key '02313012030231' and this
'mosaic_quad_key' is set to the key of a smaller internal quad, like '02313012030231300', '02313012030231301',
'023130120302313', etc, then the aforementioned '02313012030231' STAC item will be returned.

If a 'mosaic_quad_key' is set to a larger quad, like '02313012030', then the '02313012030231' quad STAC item and many
other quad STAC items that are contained by '02313012030' are returned.
        :return:
        """
        if self.stac_request.HasField("mosaic"):
            return self.stac_request.mosaic.quad_key
        return None

    @mosaic_quad_key.setter
    def mosaic_quad_key(self, value: str):
        if not self.stac_request.HasField("mosaic"):
            self.stac_request.mosaic.CopyFrom(MosaicRequest(quad_key=value))
        else:
            self.stac_request.mosaic.quad_key = value

    @property
    def offset(self) -> int:
        return self.stac_request.offset

    @offset.setter
    def offset(self, value: int):
        self.stac_request.offset = value

    @property
    def platform(self) -> enum.Platform:
        return enum.Platform(self.stac_request.platform_enum)

    @platform.setter
    def platform(self, value: enum.Platform):
        self.stac_request.platform_enum = value

    @property
    def stac_request(self) -> StacRequest:
        return self._stac_data

    @property
    def created(self) -> Optional[TimestampFilter]:
        if self.stac_request.created is None or self.stac_request.created == TimestampFilter():
            return None
        return self.stac_request.created

    @property
    def updated(self) -> Optional[TimestampFilter]:
        if self.stac_request.updated is None or self.stac_request.updated == TimestampFilter():
            return None
        return self.stac_request.updated

    @property
    def observed(self) -> Optional[TimestampFilter]:
        if self.stac_request.observed is None or self.stac_request.observed == TimestampFilter():
            return None
        return self.stac_request.observed

    def set_bounds(self, bounds: Tuple[float, float, float, float], epsg: int = 0, proj4: str = ""):
        proj = ProjectionData(proj4=proj4)
        if epsg > 0:
            proj = ProjectionData(epsg=epsg)

        bbox = EnvelopeData(xmin=bounds[0], ymin=bounds[1], xmax=bounds[2], ymax=bounds[3], proj=proj)
        self.bbox = bbox

    def set_azimuth(self,
                    rel_type: enum.FilterRelationship,
                    value: float = None,
                    start: float = None,
                    end: float = None,
                    sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED):
        if not self.stac_request.HasField("view"):
            self.stac_request.view.CopyFrom(ViewRequest())

        float_filter = StacRequestWrap.new_float_filter(rel_type, value, start, end, sort_direction)
        self.stac_request.view.azimuth.CopyFrom(float_filter)

    def set_id_complex(self,
                       value_set: Set[str],
                       rel_type: enum.FilterRelationship = enum.FilterRelationship.IN,
                       sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED):
        if len(self.id) > 0:
            self.id = ''

        self.stac_request.id_complex.CopyFrom(StringFilter(set=list(value_set),
                                                           rel_type=rel_type,
                                                           sort_direction=sort_direction))

    def set_off_nadir(self,
                      rel_type: enum.FilterRelationship,
                      value: float = None,
                      start: float = None,
                      end: float = None,
                      sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED):
        if not self.stac_request.HasField("view"):
            self.stac_request.view.CopyFrom(ViewRequest())

        float_filter = StacRequestWrap.new_float_filter(rel_type, value, start, end, sort_direction)
        self.stac_request.view.off_nadir.CopyFrom(float_filter)

    def set_sun_azimuth(self,
                        rel_type: enum.FilterRelationship,
                        value: float = None,
                        start: float = None,
                        end: float = None,
                        sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED):
        if not self.stac_request.HasField("view"):
            self.stac_request.view.CopyFrom(ViewRequest())

        float_filter = StacRequestWrap.new_float_filter(rel_type, value, start, end, sort_direction)
        self.stac_request.view.sun_azimuth.CopyFrom(float_filter)

    def set_sun_elevation(self,
                          rel_type: enum.FilterRelationship,
                          value: float = None,
                          start: float = None,
                          end: float = None,
                          sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED):
        if not self.stac_request.HasField("view"):
            self.stac_request.view.CopyFrom(ViewRequest())

        float_filter = StacRequestWrap.new_float_filter(rel_type, value, start, end, sort_direction)
        self.stac_request.view.sun_elevation.CopyFrom(float_filter)

    def set_cloud_cover(self,
                        rel_type: enum.FilterRelationship,
                        value: float = None,
                        start: float = None,
                        end: float = None,
                        sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED):
        if not self.stac_request.HasField("eo"):
            self.stac_request.eo.CopyFrom(EoRequest())

        float_filter = StacRequestWrap.new_float_filter(rel_type, value, start, end, sort_direction)
        self.stac_request.eo.cloud_cover.CopyFrom(float_filter)

    def set_gsd(self,
                rel_type: enum.FilterRelationship,
                value: float = None,
                start: float = None,
                end: float = None,
                sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED):
        float_filter = StacRequestWrap.new_float_filter(rel_type, value, start, end, sort_direction)
        self.stac_request.gsd.CopyFrom(float_filter)

    def set_observed(self,
                     rel_type: enum.FilterRelationship,
                     value: Union[datetime, date] = None,
                     start: Union[datetime, date] = None,
                     end: Union[datetime, date] = None,
                     sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED,
                     tzinfo: timezone = timezone.utc):
        self._stac_data.observed.CopyFrom(utils.pb_timestampfield(rel_type=rel_type,
                                                                  value=value,
                                                                  start=start,
                                                                  end=end,
                                                                  sort_direction=sort_direction,
                                                                  tzinfo=tzinfo))

    def merge_observed(self,
                       rel_type: enum.FilterRelationship = None,
                       value: Union[datetime, date] = None,
                       start: Union[datetime, date] = None,
                       end: Union[datetime, date] = None,
                       sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED,
                       tzinfo: timezone = timezone.utc,
                       set_if_unset=False):
        """Merges `TimestampFilter` fields into an existing `observed` field; if unset, use `set_if_unset`."""
        observed = self.observed
        if observed is None:
            if set_if_unset:
                self.set_observed(rel_type=rel_type, value=value, start=start, end=end,
                                  sort_direction=sort_direction, tzinfo=tzinfo)
        else:
            self.set_observed(rel_type=rel_type if rel_type is not None else observed.rel_type,
                              value=value if value is not None else datetime.utcfromtimestamp(observed.value.seconds),
                              start=start if start is not None else datetime.utcfromtimestamp(observed.start.seconds),
                              end=end if end is not None else datetime.utcfromtimestamp(observed.end.seconds),
                              sort_direction=sort_direction if sort_direction is not None else observed.sort_direction,
                              tzinfo=tzinfo)

    def set_created(self,
                    rel_type: enum.FilterRelationship,
                    value: Union[datetime, date] = None,
                    start: Union[datetime, date] = None,
                    end: Union[datetime, date] = None,
                    sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED,
                    tzinfo: timezone = timezone.utc):
        self._stac_data.created.CopyFrom(utils.pb_timestampfield(rel_type=rel_type,
                                                                 value=value,
                                                                 start=start,
                                                                 end=end,
                                                                 sort_direction=sort_direction,
                                                                 tzinfo=tzinfo))

    def merge_created(self,
                      rel_type: enum.FilterRelationship = None,
                      value: Union[datetime, date] = None,
                      start: Union[datetime, date] = None,
                      end: Union[datetime, date] = None,
                      sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED,
                      tzinfo: timezone = timezone.utc,
                      set_if_unset=False):
        """Merges `TimestampFilter` fields into an existing `created` field; if unset, use `set_if_unset`."""
        created = self.created
        if created is None:
            if set_if_unset:
                self.set_created(rel_type=rel_type, value=value, start=start, end=end,
                                 sort_direction=sort_direction, tzinfo=tzinfo)
        else:
            self.set_created(rel_type=rel_type if rel_type is not None else created.rel_type,
                             value=value if value is not None else datetime.utcfromtimestamp(created.value.seconds),
                             start=start if start is not None else datetime.utcfromtimestamp(created.start.seconds),
                             end=end if end is not None else datetime.utcfromtimestamp(created.end.seconds),
                             sort_direction=sort_direction if sort_direction is not None else created.sort_direction,
                             tzinfo=tzinfo)

    def set_updated(self,
                    rel_type: enum.FilterRelationship,
                    value: Union[datetime, date] = None,
                    start: Union[datetime, date] = None,
                    end: Union[datetime, date] = None,
                    sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED,
                    tzinfo: timezone = timezone.utc):
        self._stac_data.updated.CopyFrom(utils.pb_timestampfield(rel_type=rel_type,
                                                                 value=value,
                                                                 start=start,
                                                                 end=end,
                                                                 sort_direction=sort_direction,
                                                                 tzinfo=tzinfo))

    def merge_updated(self,
                      rel_type: enum.FilterRelationship = None,
                      value: Union[datetime, date] = None,
                      start: Union[datetime, date] = None,
                      end: Union[datetime, date] = None,
                      sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED,
                      tzinfo: timezone = timezone.utc,
                      set_if_unset=False):
        """Merges `TimestampFilter` fields into an existing `updated` field; if unset, use `set_if_unset`."""
        updated = self.updated
        if updated is None:
            if set_if_unset:
                self.set_updated(rel_type=rel_type, value=value, start=start, end=end,
                                 sort_direction=sort_direction, tzinfo=tzinfo)
        else:
            self.set_updated(rel_type=rel_type if rel_type is not None else updated.rel_type,
                             value=value if value is not None else datetime.utcfromtimestamp(updated.value.seconds),
                             start=start if start is not None else datetime.utcfromtimestamp(updated.start.seconds),
                             end=end if end is not None else datetime.utcfromtimestamp(updated.end.seconds),
                             sort_direction=sort_direction if sort_direction is not None else updated.sort_direction,
                             tzinfo=tzinfo)

    @staticmethod
    def new_float_filter(rel_type: enum.FilterRelationship,
                         value: float = None,
                         start: float = None,
                         end: float = None,
                         sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED):
        if value is not None:
            if start is not None or end is not None:
                raise ValueError("if value is defined, start and end cannot be used")
            elif rel_type == enum.FilterRelationship.BETWEEN or rel_type == enum.FilterRelationship.NOT_BETWEEN:
                raise ValueError("BETWEEN and NOT_BETWEEN cannot be used with value")
        else:
            if start is None or end is None:
                raise ValueError("if start is defined, end must be defined and vice versa")
            elif rel_type != enum.FilterRelationship.BETWEEN and rel_type != enum.FilterRelationship.NOT_BETWEEN:
                raise ValueError("start + end must be used with BETWEEN or NOT_BETWEEN")
        if rel_type in utils.UNSUPPORTED_TIME_FILTERS:
            raise ValueError("currently not supporting filter {}".format(rel_type.name))

        float_filter = FloatFilter(rel_type=rel_type,
                                   value=value,
                                   start=start,
                                   end=end,
                                   sort_direction=sort_direction)
        return float_filter


class CollectionRequestWrap(_BaseWrap):
    def __init__(self,
                 collection: CollectionRequest = None,
                 id: str = "",
                 correlation_id: str = None):
        if collection is None:
            collection = CollectionRequest(id=id)

        super().__init__(stac_data=collection, properties_func=None, correlation_id=correlation_id)

    @property
    def id(self) -> Optional[str]: return self.inner.id

    @id.setter
    def id(self, value: str): self.inner.id = value

    @property
    def bbox(self) -> EnvelopeData:
        if self.inner.HasField("bbox"):
            return self.inner.bbox
        elif self.inner.HasField("intersects"):
            # TODO mono-759, this seems messy
            return _to_envelope_data(_from_protobuf(self.intersects))
        return None

    @bbox.setter
    def bbox(self, value: EnvelopeData):
        # this tests the spatial reference (it would be better to have a dedicated method)
        self.inner.bbox.CopyFrom(value)
        self.inner.ClearField("intersects")
        # TODO mono-759, this seems messy
        self.inner.intersects.CopyFrom(_to_protobuf(_from_envelope_data(value), proj=value.proj))

    @property
    def intersects(self) -> Optional[BaseGeometry]:
        if self.inner.HasField("intersects"):
            return _from_protobuf(self.inner.intersects)
        elif self.inner.HasField("bbox"):
            return _from_envelope_data(self.bbox)
        return None

    @intersects.setter
    def intersects(self, value: BaseGeometry, proj: ProjectionData = None):
        if isinstance(value, BaseGeometry):
            if proj is None:
                print("warning, no projection data set. assuming WGS84")
                proj = ProjectionData(epsg=4326)
            g = GeometryData(wkb=value.wkb, proj=proj)
            self.inner.intersects.CopyFrom(g)
        else:
            # assume epl.geometry
            self.inner.intersects.CopyFrom(value.geometry_data)
        self.inner.ClearField("bbox")

    @property
    def intersects_proj(self) -> ProjectionData:
        if self.inner.HasField("intersects"):
            return self.inner.intersects.proj
        elif self.inner.HasField("bbox"):
            return self.inner.bbox.proj
        return None

    @intersects_proj.setter
    def intersects_proj(self, value: ProjectionData):
        if self.inner.HasField("geometry"):
            self.inner.intersects.proj.CopyFrom(value)
            return
        raise ValueError("can't set intersects projection without first setting geometry")

    @property
    def observed_start(self) -> Optional[TimestampFilter]:
        if self.inner.observed_start is None or \
                self.inner.observed_start == TimestampFilter():
            return None
        return self.inner.observed_start

    @property
    def observed_end(self) -> Optional[TimestampFilter]:
        if self.inner.observed_end is None or self.inner.observed_end == TimestampFilter():
            return None
        return self.inner.observed_end

    @property
    def last_updated(self) -> Optional[TimestampFilter]:
        if self.inner.last_updated is None or self.inner.last_updated == TimestampFilter():
            return None
        return self.inner.last_updated

    def set_observed_start(self,
                           rel_type: enum.FilterRelationship,
                           value: Union[datetime, date] = None,
                           start: Union[datetime, date] = None,
                           end: Union[datetime, date] = None,
                           sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED,
                           tzinfo: timezone = timezone.utc):
        self.inner.observed_start.CopyFrom(utils.pb_timestampfield(rel_type=rel_type,
                                                                   value=value,
                                                                   start=start,
                                                                   end=end,
                                                                   sort_direction=sort_direction,
                                                                   tzinfo=tzinfo))

    def set_observed_end(self,
                         rel_type: enum.FilterRelationship,
                         value: Union[datetime, date] = None,
                         start: Union[datetime, date] = None,
                         end: Union[datetime, date] = None,
                         sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED,
                         tzinfo: timezone = timezone.utc):
        self.inner.observed_end.CopyFrom(utils.pb_timestampfield(rel_type=rel_type,
                                                                 value=value,
                                                                 start=start,
                                                                 end=end,
                                                                 sort_direction=sort_direction,
                                                                 tzinfo=tzinfo))

    def set_last_updated(self,
                         rel_type: enum.FilterRelationship,
                         value: Union[datetime, date] = None,
                         start: Union[datetime, date] = None,
                         end: Union[datetime, date] = None,
                         sort_direction: enum.SortDirection = enum.SortDirection.NOT_SORTED,
                         tzinfo: timezone = timezone.utc):
        self.inner.last_updated.CopyFrom(utils.pb_timestampfield(rel_type=rel_type,
                                                                 value=value,
                                                                 start=start,
                                                                 end=end,
                                                                 sort_direction=sort_direction,
                                                                 tzinfo=tzinfo))

    @property
    def is_static(self) -> Optional[bool]: return self.inner.is_static

    @is_static.setter
    def is_static(self, value: bool): self.inner.is_static = value

    @property
    def inner(self) -> CollectionRequest: return self._stac_data


class CollectionWrap(_BaseWrap):
    def __init__(self, collection: Collection = None, correlation_id: str = None):
        collection_data = Collection()
        collection_data.extent.CopyFrom(Extent())
        if collection is not None:
            collection_data.CopyFrom(collection)

        super().__init__(stac_data=collection_data, properties_func=None, correlation_id=correlation_id)

    @property
    def id(self) -> str: return self.inner.id

    @id.setter
    def id(self, value: str): self.inner.id = value

    @property
    def title(self) -> str: return self.inner.title

    @title.setter
    def title(self, value: str): self.inner.title = value

    @property
    def description(self) -> str: return self.inner.description

    @description.setter
    def description(self, value: str): self.inner.description = value

    @property
    def keywords(self) -> List[str]: return self.inner.keywords

    @keywords.setter
    def keywords(self, value: List[str]): self.inner.keywords[:] = value

    @property
    def license(self) -> str: return self.inner.license

    @license.setter
    def license(self, value: str): self.inner.license = value

    @property
    def provider(self) -> Optional[Provider]:
        if len(self.inner.providers) == 0:
            return None
        return self.inner.providers[0]

    @provider.setter
    def provider(self, value: Provider):
        del self.inner.providers[:]
        self.inner.providers.append(value)

    @property
    def bbox(self) -> Optional[EnvelopeData]:
        if self.inner.extent is None or len(self.inner.extent.spatial) == 0:
            return None
        return self.inner.extent.spatial[0]

    @bbox.setter
    def bbox(self, value: EnvelopeData):
        if self.inner.extent is None:
            self.inner.extent = Extent()
        del self.inner.extent.spatial[:]
        self.inner.extent.spatial.append(value)

    @property
    def observed(self) -> Optional[Tuple[datetime, datetime]]:
        if self.inner.extent is None or len(self.inner.extent.temporal) == 0:
            return None
        interval = self.inner.extent.temporal[0]
        if interval.start is None or interval.end is None:
            return None
        return utils.datetime_from_pb_timestamp(interval.start), utils.datetime_from_pb_timestamp(interval.end)

    @observed.setter
    def observed(self, value: Tuple[datetime, datetime]):
        if self.inner.extent is None:
            self.inner.extent = Extent()
        interval = Interval(start=utils.pb_timestamp(value[0]), end=utils.pb_timestamp(value[1]))
        del self.inner.extent.temporal[:]
        self.inner.extent.temporal.append(interval)

    @property
    def footprint(self) -> Optional[BaseGeometry]:
        if self.inner.extent is None:
            return None
        if self.inner.extent.footprint is None:
            return None
        return _from_protobuf(self.inner.extent.footprint)

    @footprint.setter
    def footprint(self, value: BaseGeometry, proj: ProjectionData = None):
        if self.inner.extent is None:
            self.inner.extent = Extent()
        if isinstance(value, BaseGeometry):
            if proj is None:
                print("warning, no projection data set. assuming WGS84")
                proj = ProjectionData(epsg=4326)
            g = GeometryData(wkb=value.wkb, proj=proj)
            self.inner.extent.intersects.CopyFrom(g)
        else:
            # assume epl.geometry
            self.inner.extent.intersects.CopyFrom(value.geometry_data)

    # TODO: summaries

    @property
    def is_static(self) -> bool: return self.inner.is_static

    @is_static.setter
    def is_static(self, value: bool): self.inner.is_static = value

    @property
    def stac_request(self) -> Optional[StacRequestWrap]:
        if self.inner.stac_request is None:
            return None
        return StacRequestWrap(self.inner.stac_request)

    @stac_request.setter
    def stac_request(self, value: StacRequestWrap): self.inner.stac_request.CopyFrom(value.stac_request)

    @property
    def inner(self) -> Collection: return self._stac_data


class NSLClientEx(NSLClient):
    def __init__(self, nsl_only=False, **kwargs):
        super().__init__(nsl_only=nsl_only, **kwargs)
        self._internal_stac_service = stac_singleton

    def update_service_url(self, stac_service_url):
        """
        update the stac service address
        :param stac_service_url: localhost:8080, 34.34.34.34:9000, http://demo.nearspacelabs.com, etc
        :return:
        """
        super().update_service_url(stac_service_url)
        self._internal_stac_service.update_service_url(stac_service_url=stac_service_url)

    def search_ex(self,
                  stac_request_wrapped: StacRequestWrap,
                  timeout=15,
                  nsl_id: str = None,
                  profile_name: str = None,
                  auto_paginate: bool = False,
                  only_accessible: bool = False,
                  page_size: int = 50) -> Iterator[StacItemWrap]:
        for stac_item in self.search(stac_request_wrapped.stac_request,
                                     timeout=timeout,
                                     nsl_id=nsl_id,
                                     profile_name=profile_name,
                                     auto_paginate=auto_paginate,
                                     only_accessible=only_accessible,
                                     page_size=page_size,
                                     correlation_id=stac_request_wrapped.correlation_id):
            yield StacItemWrap(stac_item=stac_item)

    def feature_collection_ex(self,
                              stac_request_wrapped: StacRequestWrap,
                              timeout=15,
                              nsl_id: str = None,
                              profile_name: str = None,
                              feature_collection: Dict = None,
                              auto_paginate: bool = False) -> Dict:
        if feature_collection is None:
            feature_collection = {'type': 'FeatureCollection', 'features': []}

        for item in self.search_ex(stac_request_wrapped,
                                   timeout=timeout,
                                   nsl_id=nsl_id,
                                   profile_name=profile_name,
                                   auto_paginate=auto_paginate):
            feature_collection['features'].append(item.feature)

        return feature_collection

    def search_one_ex(self,
                      stac_request_wrapped: StacRequestWrap,
                      timeout=15,
                      nsl_id: str = None,
                      profile_name: str = None) -> Optional[StacItemWrap]:
        stac_item = self.search_one(stac_request=stac_request_wrapped.stac_request,
                                    timeout=timeout, nsl_id=nsl_id, profile_name=profile_name,
                                    correlation_id=stac_request_wrapped.correlation_id)
        if not stac_item.id:
            return None
        return StacItemWrap(stac_item=stac_item)

    def count_ex(self,
                 stac_request_wrapped: StacRequestWrap,
                 timeout=15,
                 nsl_id: str = None,
                 profile_name: str = None) -> int:
        return self.count(stac_request=stac_request_wrapped.stac_request,
                          timeout=timeout, nsl_id=nsl_id, profile_name=profile_name,
                          correlation_id=stac_request_wrapped.correlation_id)

    def search_collections_ex(self,
                              collection_request: CollectionRequestWrap,
                              timeout=15,
                              nsl_id: str = None,
                              profile_name: str = None) -> Iterator[CollectionWrap]:
        for collection in self.search_collections(collection_request.inner,
                                                  timeout=timeout,
                                                  nsl_id=nsl_id,
                                                  profile_name=profile_name,
                                                  correlation_id=collection_request.correlation_id):
            yield CollectionWrap(collection=collection)

    def subscribe_ex(self,
                     stac_request_wrapped: StacRequestWrap,
                     destination: BaseDestination,
                     nsl_id: str = None,
                     profile_name: str = None) -> str:
        return self.subscribe(stac_request=stac_request_wrapped.stac_request, destination=destination,
                              nsl_id=nsl_id, profile_name=profile_name)

    def resubscribe_ex(self, sub_id: str, nsl_id: str = None, profile_name: str = None):
        return self.resubscribe(sub_id=sub_id, nsl_id=nsl_id, profile_name=profile_name)

    def unsubscribe_ex(self, sub_id: str, nsl_id: str = None, profile_name: str = None):
        return self.unsubscribe(sub_id=sub_id, nsl_id=nsl_id, profile_name=profile_name)

    def subscriptions_ex(self, nsl_id: str = None, profile_name: str = None) -> List[Subscription]:
        return self.subscriptions(nsl_id=nsl_id, profile_name=profile_name)
